import json
import pathlib
from ctypes import *
from os import pathsep, environ
from os.path import exists, abspath, join
from typing import Dict, Tuple, Optional

dll_names = {
    "12c": ("frmd2f.dll", "msvcr100"),
    "10g": ("frmd2f.dll", "msvcrt"),
    "6i": ("ifd2f60.dll", "msvcrt"),
}


def find_dll(dll_name: str) -> Optional[str]:
    search_path = environ["PATH"]
    for path in search_path.split(pathsep):
        if exists(join(path, dll_name)):
            return abspath(path)
    return None


def dlls(version: str) -> Tuple[CDLL, CDLL]:
    api_dll, cdll_name = dll_names[version]
    dll_path = find_dll(api_dll)
    if dll_path:
        from os import add_dll_directory

        msvcrt = cdll.LoadLibrary(cdll_name)
        with add_dll_directory(dll_path):
            return cdll.LoadLibrary(api_dll), msvcrt
    raise ImportError("No Oracle Forms API found")


def read_api_objects(version: str) -> Dict:  # type: ignore
    file_path = pathlib.Path(__file__).parent / "forms_api" / f"parsed_{version}.json"
    with open(file_path, mode="r", encoding="utf-8") as file:
        json_data: Dict = json.load(file)  # type: ignore
    return json_data



from typing import NoReturn

error_mapping = {
    0: "Operation Succeeded",
    1: "Operation Failed",
    2: "Operation returned YES",
    3: "Operation returned NO",
    4: "Bad context provided",
    5: "Bad property",
    6: "One of the args is wrong",
    7: "Object type is unknown",
    8: "Unexpected object",
    9: "Unexpected parent",
    10: "Null object passed",
    11: "Null pointer to object",
    12: "Null property",
    13: "Not connected to DB",
    14: "Out of memory",
    15: "Message file not found",
    16: "Generation failed",
    17: "Not implemented yet",
    18: "Passed in type does not match the actual object",
    19: "The operation failed partially, but the error was not fatal",
    20: "Null Data Passed in",
    21: "Data Passed in is invalid",
    22: "Index in is invalid",
    23: "The Object does not have the given property",
    24: "Initialization failed because a NULL Instance Handle passed (Applies to Windows only)",
    25: "The operation failed because the object being created/placed was not unique. "
    "An object with this name already exists",
    26: "The Object was not found",
    27: "Function can only be called in translation mode",
    28: "The database context passed in is invalid",
    29: "A subclassed module could not be found during loading",
    30: "Duplicate String ID (This slot in the string table has already been used).",
    31: "A supplied value parameter was out of the legal range",
    32: "The specified file was not found",
    33: "An attached library could not be found",
}


class FormsException(Exception):
    pass


def raise_for_code(error_code: int) -> NoReturn:
    raise FormsException(error_code, f"{error_mapping[error_code]}")



from __future__ import annotations

from ctypes import c_void_p
from types import TracebackType
from typing import Optional, Type, Union, Any, List, TypeVar

from .context import create_module
from .context import load_module
from .context import save_module
from .generic_object import GenericObject, BaseObject, FormsObjects
from .misc import (
    forms_object,
    Text,
    Number,
    Object,
    Bool,
    Subobjects,
    Unknown,
    Constant,
)
from .constants import *


# satisfy both MyPy and PyCharm IDE autocomplete
U = TypeVar("U")
ObjectList = Union[List[U], Subobjects[U]]
Obj = Union[U, Object[U]]


@forms_object
class Module(BaseObject):
    object_type = FormsObjects.module

    name = Text("NAME")
    comments = Text("COMMENT")
    console_window = Text("CONSOLE_WIN")
    cursor_mode: Constant[CursorMode] = Constant("CRSR_MODE", CursorMode)
    direction: Constant[Direction] = Constant("LANG_DIR", Direction)
    first_navigation_data_block = Text("FRST_NAVIGATION_BLK_NAM")
    menu_module = Text("MNU_MOD")
    menu_role = Text("MNU_ROLE")
    initial_menu = Text("INIT_MNU")
    form_horizontal_toolbar_canvas = Text("HORZ_TLBR_CNV")
    form_vertical_toolbar_canvas = Text("VERT_TLBR_CNV")
    mouse_navigation_limit: Constant[MouseNavigationLimit] = Constant(
        "MOUSE_NAVIGATION_LMT", MouseNavigationLimit
    )
    current_record_visual_attribute_group = Text("REC_VAT_GRP_NAM")
    savepoint_mode = Bool("SVPNT_MODE")
    title = Text("TITLE")
    use_3d_controls = Bool("USE_3D_CNTRLS")
    validation_unit: Constant[ValidationUnit] = Constant(
        "VALIDATION_UNIT", ValidationUnit
    )
    title_string_id = Number("TITLE_STRID")
    interaction_mode: Constant[InteractionMode] = Constant(
        "INTERACTION_MODE", InteractionMode
    )
    maximum_query_time = Number("MAX_QRY_TIME")
    maximum_records_fetched = Number("MAX_RECS_FETCHED")
    isolation_mode: Constant[IsolationMode] = Constant("ISOLATION_MODE", IsolationMode)
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    runtime_compatibility_mode: Constant[RuntimeCompatibilityMode] = Constant(
        "RUNTIME_COMP", RuntimeCompatibilityMode
    )
    parent_objects_type = Number("PAR_TYP")
    help_book_title = Text("HELP_BOOK_TITLE")
    defer_required_enforcement = Number("NEWDEFER_REQ_ENF")

    alerts: ObjectList[Alert] = Subobjects("ALERT")
    attached_libraries: ObjectList[AttachedLibrary] = Subobjects("ATT_LIB")
    data_blocks: ObjectList[DataBlock] = Subobjects("BLOCK")
    canvases: ObjectList[Canvas] = Subobjects("CANVAS")
    editors: ObjectList[Editor] = Subobjects("EDITOR")
    form_parameters: ObjectList[FormParameter] = Subobjects("FORM_PARAM")
    lovs: ObjectList[LOV] = Subobjects("LOV")
    menus: ObjectList[Menu] = Subobjects("MENU")
    triggers: ObjectList[Trigger] = Subobjects("TRIGGER")
    visual_attributes: ObjectList[VisualAttribute] = Subobjects("VIS_ATTR")
    windows: ObjectList[Window] = Subobjects("WINDOW")
    reports: ObjectList[Report] = Subobjects("REPORT")
    object_groups: ObjectList[ObjectGroup] = Subobjects("OBJ_GRP")
    program_units: ObjectList[ProgramUnit] = Subobjects("PROG_UNIT")
    property_classes: ObjectList[PropertyClass] = Subobjects("PROP_CLASS")
    record_groups: ObjectList[RecordGroup] = Subobjects("REC_GRP")
    events: ObjectList[Event] = Subobjects("EVENT")

    next_object: Obj[Module] = Object("NEXT")
    previous_object: Obj[Module] = Object("PREVIOUS")

    first_data_block_object: Obj[DataBlock] = Object("FRST_NAVIGATION_BLK_OBJ")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    current_record_va_pointer: Obj[VisualAttribute] = Object("REC_VAT_GRP_OBJ")

    def __init__(self, module: Union[c_void_p, BaseObject], path: Optional[str] = None):
        super().__init__(module)
        self.path = path

    def __enter__(self) -> Module:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.destroy()

    @classmethod
    def create(cls, module_name: str) -> Module:
        return cls(create_module(module_name), module_name)

    @classmethod
    def load(cls, path: str) -> Module:
        return cls(load_module(path), path=path)

    def save(self, path: Optional[str] = None) -> None:
        path = path or self.path
        if path is None:
            raise ValueError("No path provided or available")
        save_module(module=self, path=path)


@forms_object
class AttachedLibrary(GenericObject):
    # auto-generated
    object_type = FormsObjects.attached_library

    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    pl_sql_library_location = Text("LIB_LOC")
    pl_sql_library_source = Number("LIB_SRC")
    name = Text("NAME")
    next_object: Obj[AttachedLibrary] = Object("NEXT")
    previous_object: Obj[AttachedLibrary] = Object("PREVIOUS")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class Alert(GenericObject):
    # auto-generated
    object_type = FormsObjects.alert

    button_1_label = Text("BTN_1_LBL")
    button_2_label = Text("BTN_2_LBL")
    button_3_label = Text("BTN_3_LBL")
    default_alert_button: Constant[DefaultAlertButton] = Constant(
        "DFLT_ALT_BTN", DefaultAlertButton
    )
    message = Text("ALT_MSG")
    alert_style: Constant[AlertStyle] = Constant("ALT_STY", AlertStyle)
    background_color = Text("BACK_COLOR")
    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    direction: Constant[Direction] = Constant("LANG_DIR", Direction)
    fill_pattern = Text("FILL_PAT")
    font_name = Text("FONT_NAM")
    font_size = Number("FONT_SIZ")
    font_style: Constant[FontStyle] = Constant("FONT_STY", FontStyle)
    font_weight: Constant[FontWeight] = Constant("FONT_WGHT", FontWeight)
    font_spacing: Constant[FontSpacing] = Constant("FONT_SPCING", FontSpacing)
    foreground_color = Text("FORE_COLOR")
    name = Text("NAME")
    next_object: Obj[Alert] = Object("NEXT")
    title = Text("TITLE")
    visual_attribute_group = Text("VAT_NAM")
    va_object: Obj[VisualAttribute] = Object("VAT_OBJ")
    previous_object: Obj[Alert] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class Canvas(GenericObject):
    # auto-generated
    object_type = FormsObjects.canvas

    background_color = Text("BACK_COLOR")
    bevel: Constant[Bevel] = Constant("BEVEL", Bevel)
    graphics: ObjectList[Graphic] = Subobjects("GRAPHIC")
    case_info = Unknown("CLIENT_INFO")
    canvas_type: Constant[CanvasType] = Constant("CNV_TYP", CanvasType)
    comments = Text("COMMENT")
    direction: Constant[Direction] = Constant("LANG_DIR", Direction)
    visible = Bool("VISIBLE")
    display_viewport = Bool("DISP_VIEWPORT")
    viewport_x_position = Number("VPRT_X_POS")
    viewport_y_position = Number("VPRT_Y_POS")
    fill_pattern = Text("FILL_PAT")
    font_name = Text("FONT_NAM")
    font_size = Number("FONT_SIZ")
    font_style: Constant[FontStyle] = Constant("FONT_STY", FontStyle)
    font_weight: Constant[FontWeight] = Constant("FONT_WGHT", FontWeight)
    font_spacing: Constant[FontSpacing] = Constant("FONT_SPCING", FontSpacing)
    foreground_color = Text("FORE_COLOR")
    height = Number("HEIGHT")
    show_horizontal_scroll_bar = Bool("SHOW_HORZ_SCRLBR")
    name = Text("NAME")
    next_object: Obj[Canvas] = Object("NEXT")
    raise_on_entry = Bool("RAISE_ON_ENT")
    visual_attribute_group = Text("VAT_NAM")
    va_object: Obj[VisualAttribute] = Object("VAT_OBJ")
    show_vertical_scroll_bar = Bool("SHOW_VERT_SCRLBR")
    viewport_height = Number("VPRT_HGT")
    viewport_width = Number("VPRT_WID")
    viewport_x_position_on_canvas = Number("VPRT_X_POS_ON_CNV")
    viewport_y_position_on_canvas = Number("VPRT_Y_POS_ON_CNV")
    visual_stat: Obj[VisualState] = Object("VIS_STATE")
    width = Number("WIDTH")
    window = Text("WND_NAM")
    window_object_pointer: Obj[Window] = Object("WND_OBJ")
    previous_object: Obj[Canvas] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    tab_pages: ObjectList[TabPage] = Subobjects("TAB_PAGE")
    popup_menu = Text("POPUP_MNU_NAM")
    popup_menu_object: Obj[BaseObject] = Object("POPUP_MNU_OBJ")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    tab_attachment_edge: Constant[TabAttachmentEdge] = Constant(
        "TAB_ATT_EDGE", TabAttachmentEdge
    )
    corner_style: Constant[CornerStyle] = Constant("TAB_STY", CornerStyle)
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_type = Number("PAR_TYP")
    help_book_topic = Text("HELP_BOOK_TOPIC")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")
    active_style = Number("TAB_ACT_STY")
    width_style = Number("TAB_WID_STY")
    gradient_start_side = Number("GRADIENT_START")


# todo: todo: undocumented, no .h file
@forms_object
class CompoundText(GenericObject):
    # auto-generated
    object_type = FormsObjects.compound_text


@forms_object
class DataBlock(GenericObject):
    # auto-generated
    object_type = FormsObjects.data_block

    background_color = Text("BACK_COLOR")
    case_info = Unknown("CLIENT_INFO")
    enforce_column_security = Bool("ENFRC_COL_SECURITY")
    comments = Text("COMMENT")
    delete_allowed = Bool("DEL_ALLOWED")
    direction: Constant[Direction] = Constant("LANG_DIR", Direction)
    fill_pattern = Text("FILL_PAT")
    foreground_color = Text("FORE_COLOR")
    insert_allowed = Bool("INSRT_ALLOWED")
    items: ObjectList[Item] = Subobjects("ITEM")
    key_mode: Constant[KeyMode] = Constant("KEY_MODE", KeyMode)
    locking_mode: Constant[LockingMode] = Constant("LOCK_MODE", LockingMode)
    name = Text("NAME")
    navigation_style: Constant[NavigationStyle] = Constant(
        "NAVIGATION_STY", NavigationStyle
    )
    next_object: Obj[DataBlock] = Object("NEXT")
    next_navigation_data_block = Text("NXT_NAVIGATION_BLK_NAM")
    next_data_block_object: Obj[DataBlock] = Object("NXT_NAVIGATION_BLK_OBJ")
    optimizer_hint = Text("OPT_HINT")
    order_by_clause = Text("ORDR_BY_CLAUSE")
    previous_navigation_data_block = Text("PREV_NAVIGATION_BLK_NAM")
    previous_data_block_object: Obj[DataBlock] = Object("PREV_NAVIGATION_BLK_OBJ")
    query_allowed = Bool("QRY_ALLOWED")
    number_of_records_buffered = Number("RECS_BUFFERED_COUNT")
    number_of_records_displayed = Number("RECS_DISP_COUNT")
    query_array_size = Number("RECS_FETCHED_COUNT")
    record_orientation: Constant[RecordOrientation] = Constant(
        "REC_ORNT", RecordOrientation
    )
    current_record_visual_attribute_group = Text("REC_VAT_GRP_NAM")
    relation_object: Obj[Relation] = Object("REL")
    reverse_direction = Bool("REV_DIR")
    scroll_bar_canvas = Text("SCRLBR_CNV_NAM")
    scroll_bar_orientation: Constant[ScrollBarOrientation] = Constant(
        "SCRLBR_ORNT", ScrollBarOrientation
    )
    scroll_bar_width = Number("SCRLBR_WID")
    scroll_bar_x_position = Number("SCRLBR_X_POS")
    scroll_bar_y_position = Number("SCRLBR_Y_POS")
    show_scroll_bar = Bool("SHOW_SCRLBR")
    triggers: ObjectList[Trigger] = Subobjects("TRIGGER")
    update_allowed = Bool("UPDT_ALLOWED")
    update_changed_columns_only = Bool("UPDT_CHANGED_COLS")
    visual_attribute_group = Text("VAT_NAM")
    va_object: Obj[VisualAttribute] = Object("VAT_OBJ")
    where_clause = Text("WHERE_CLAUSE")
    enforce_primary_key = Bool("ENFRC_PRMRY_KEY")
    current_record_va_pointer: Obj[VisualAttribute] = Object("REC_VAT_GRP_OBJ")
    previous_object: Obj[DataBlock] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    scrollbar_canvas_object_pointer: Obj[Canvas] = Object("SCRLBR_CNV_OBJ")
    dml_data_target_name = Text("DML_DAT_NAM")
    query_data_source_name = Text("QRY_DAT_SRC_NAM")
    database_data_block = Bool("DB_BLK")
    query_data_source_type: Constant[QueryDataSourceType] = Constant(
        "QRY_DAT_SRC_TYP", QueryDataSourceType
    )
    query_all_records = Bool("QRY_ALL_RECS")
    dml_data_target_type: Constant[DMLDataTargetType] = Constant(
        "DML_DAT_TYP", DMLDataTargetType
    )
    dml_array_size = Number("DML_ARY_SIZ")
    insert_procedure_name = Text("INSRT_PROC_NAM")
    update_procedure_name = Text("UPDT_PROC_NAM")
    delete_procedure_name = Text("DEL_PROC_NAM")
    lock_procedure_name = Text("LOCK_PROC_NAM")
    single_record = Bool("SNGL_REC")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    scroll_bar_tab_page = Text("SCRLBR_TBP_NAM")
    scrollbar_tab_object: Obj[TabPage] = Object("SCRLBR_TBP_OBJ")
    maximum_query_time = Number("MAX_QRY_TIME")
    maximum_records_fetched = Number("MAX_RECS_FETCHED")
    precompute_summaries = Bool("PRECOMP_SUMM")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    query_column_object: Obj[DataSourceColumn] = Object("QRY_DAT_SRC_COL")
    query_argument_object: Obj[DataSourceArgument] = Object("QRY_DAT_SRC_ARG")
    insert_column_object: Obj[DataSourceColumn] = Object("INS_DAT_SRC_COL")
    insert_argument_object: Obj[DataSourceArgument] = Object("INS_DAT_SRC_ARG")
    update_column_object: Obj[DataSourceColumn] = Object("UPD_DAT_SRC_COL")
    update_argument_object: Obj[DataSourceArgument] = Object("UPD_DAT_SRC_ARG")
    delete_column_object: Obj[DataSourceColumn] = Object("DEL_DAT_SRC_COL")
    delete_argument_object: Obj[DataSourceArgument] = Object("DEL_DAT_SRC_ARG")
    lock_column_object: Obj[DataSourceColumn] = Object("LOCK_DAT_SRC_COL")
    lock_argument_object: Obj[DataSourceArgument] = Object("LOCK_DAT_SRC_ARG")
    parent_objects_type = Number("PAR_TYP")
    scroll_bar_length = Number("SCRLBR_LEN")
    alias = Text("ALIAS")
    include_ref_item = Bool("INCLUDE_REFITEM")
    dml_returning_value = Bool("DML_RET_VAL")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")
    row_banding_frequency = Number("ROW_BANDING_FREQ")


@forms_object
class FormParameter(GenericObject):
    # auto-generated
    object_type = FormsObjects.form_parameter

    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    parameter_data_type: Constant[ParameterDataType] = Constant(
        "PARAM_DAT_TYP", ParameterDataType
    )
    parameter_initial_value = Text("PARAM_INIT_VAL")
    name = Text("NAME")
    next_object: Obj[FormParameter] = Object("NEXT")
    previous_object: Obj[FormParameter] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class Graphic(GenericObject):
    # auto-generated
    object_type = FormsObjects.graphic

    background_color = Text("BACK_COLOR")
    bevel: Constant[Bevel] = Constant("BEVEL", Bevel)
    graphics: ObjectList[Graphic] = Subobjects("GRAPHIC")
    case_info = Unknown("CLIENT_INFO")
    image_format: Constant[ImageFormat] = Constant("IMG_FMT", ImageFormat)
    image_depth: Constant[ImageDepth] = Constant("IMG_DPTH", ImageDepth)
    direction: Constant[Direction] = Constant("LANG_DIR", Direction)
    fill_pattern = Text("FILL_PAT")
    foreground_color = Text("FORE_COLOR")
    height = Number("HEIGHT")
    name = Text("NAME")
    next_object: Obj[Graphic] = Object("NEXT")
    display_quality: Constant[DisplayQuality] = Constant("DISP_QLTY", DisplayQuality)
    number_of_records_displayed = Number("RECS_DISP_COUNT")
    distance_between_records = Number("DIST_BTWN_RECS")
    scroll_bar_width = Number("SCRLBR_WID")
    show_scroll_bar = Bool("SHOW_SCRLBR")
    visual_attribute_group = Text("VAT_NAM")
    va_object: Obj[VisualAttribute] = Object("VAT_OBJ")
    width = Number("WIDTH")
    x_position = Number("X_POS")
    y_position = Number("Y_POS")
    graphics_type: Constant[GraphicsType] = Constant("GRAPHICS_TYP", GraphicsType)
    closed = Bool("CLOSED")
    arrow_style: Constant[ArrowStyle] = Constant("ARROW_STY", ArrowStyle)
    x_corner_radius = Number("CORNER_RADIUS_X")
    y_corner_radius = Number("CORNER_RADIUS_Y")
    start_angle = Number("INTERNAL_STRT_ANGLE")
    end_angle = Number("INTERNAL_END_ANGLE")
    previous_object: Obj[Graphic] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    dither = Bool("DITHER")
    clip_x_position = Number("CLIP_X_POS")
    clip_y_position = Number("CLIP_Y_POS")
    clip_width = Number("CLIP_WID")
    clip_height = Number("CLIP_HGT")
    points: ObjectList[Point] = Subobjects("POINT")
    frame_title_reading_order: Constant[ReadingOrder] = Constant(
        "TTL_READING_ORDR", ReadingOrder
    )
    frame_title_alignment: Constant[FrameTitleAlignment] = Constant(
        "FRAME_TTL_ALIGN", FrameTitleAlignment
    )
    frame_title_offset = Number("FRAME_TTL_OFST")
    frame_title_spacing = Number("FRAME_TTL_SPCING")
    frame_title = Text("FRAME_TTL")
    layout_style: Constant[LayoutStyle] = Constant("LAYOUT_STY", LayoutStyle)
    frame_alignment: Constant[FrameAlignment] = Constant("FRAME_ALIGN", FrameAlignment)
    single_object_alignment: Constant[SingleObjectAlignment] = Constant(
        "SNGL_OBJ_ALIGN", SingleObjectAlignment
    )
    vertical_fill = Bool("VERT_FILL")
    maximum_objects_per_line = Number("MAX_OBJS")
    allow_expansion = Bool("ALLOW_EXPANSION")
    horizontal_margin = Number("HORZ_MARGN")
    vertical_margin = Number("VERT_MARGN")
    vertical_object_offset = Number("VERT_OBJ_OFST")
    horizontal_object_offset = Number("HORZ_OBJ_OFST")
    start_prompt_offset = Number("STRT_PRMPT_OFST")
    top_prompt_offset = Number("TOP_PRMPT_OFST")
    start_prompt_alignment: Constant[PromptAlignment] = Constant(
        "STRT_PRMPT_ALIGN", PromptAlignment
    )
    top_prompt_alignment: Constant[PromptAlignment] = Constant(
        "TOP_PRMPT_ALIGN", PromptAlignment
    )
    allow_multi_line_prompts = Bool("ALLOW_MLT_LIN_PRMPTS")
    allow_top_attached_prompts = Bool("ALLOW_TOP_ATT_PRMPTS")
    allow_start_attached_prompts = Bool("ALLOW_STRT_ATT_PRMPTS")
    update_layout: Constant[UpdateLayout] = Constant("UPDT_LAYOUT", UpdateLayout)
    line_width = Number("INTERNAL_LIN_WID")
    rotation_angle = Number("INTERNAL_ROTATION_ANGLE")
    edge_foreground_color = Text("EDGE_FORE_COLOR")
    edge_background_color = Text("EDGE_BACK_COLOR")
    edge_pattern = Text("EDGE_PAT")
    dash_style: Constant[DashStyle] = Constant("DASH_STY", DashStyle)
    cap_style: Constant[CapStyle] = Constant("CAP_STY", CapStyle)
    join_style: Constant[JoinStyle] = Constant("JOIN_STY", JoinStyle)
    line_spacing: Constant[LineSpacing] = Constant("LIN_SPCING", LineSpacing)
    custom_spacing = Number("CSTM_SPCING")
    fixed_bounding_box = Bool("FIXED_BOUNDING_BX")
    wrap_text = Bool("WRAP_TXT")
    bounding_box_scaleable = Bool("BOUNDING_BX_SCALABLE")
    font_scaleable = Bool("FONT_SCALEABLE")
    horizontal_origin: Constant[HorizontalOrigin] = Constant(
        "HORZ_ORGN", HorizontalOrigin
    )
    vertical_origin: Constant[VerticalOrigin] = Constant("VERT_ORGN", VerticalOrigin)
    horizontal_justification: Constant[Justification] = Constant(
        "HORZ_JST", Justification
    )
    vertical_justification: Constant[Justification] = Constant(
        "VERT_JST", Justification
    )
    compound_text_object: Obj[CompoundText] = Object("CMPTXT")
    tab_page = Text("TBP_NAM")
    tab_page_name: Obj[BaseObject] = Object("TBP_OBJ")
    layout_data_block = Text("LAYOUT_DATA_BLK_NAM")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    title_va_object: Obj[VisualAttribute] = Object("FRAME_TTL_VAT_OBJ")
    frame_title_visual_attribute_group = Text("FRAME_TTL_VAT_NAM")
    frame_title_foreground_color = Text("FRAME_TTL_FORE_COLOR")
    frame_title_background_color = Text("FRAME_TTL_BACK_COLOR")
    frame_title_fill_pattern = Text("FRAME_TTL_FILL_PAT")
    frame_title_font_name = Text("FRAME_TTL_FONT_NAM")
    frame_title_font_size = Number("FRAME_TTL_FONT_SIZ")
    frame_title_font_style: Constant[FontStyle] = Constant(
        "FRAME_TTL_FONT_STY", FontStyle
    )
    frame_title_font_weight: Constant[FontWeight] = Constant(
        "FRAME_TTL_FONT_WGHT", FontWeight
    )
    frame_title_font_spacing: Constant[FontSpacing] = Constant(
        "FRAME_TTL_FONT_SPCING", FontSpacing
    )
    scroll_bar_alignment: Constant[ScrollBarAlignment] = Constant(
        "SCRLBR_ALIGN", ScrollBarAlignment
    )
    shrinkwrap = Bool("SHRINKWRAP")
    graphic_object_font_name = Text("GRA_FONT_NAM")
    graphic_object_font_size = Number("GRA_FONT_SIZ")
    graphic_object_font_style: Constant[FontStyle] = Constant("GRA_FONT_STY", FontStyle)
    graphic_object_text = Text("GRA_TEXT")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    graphic_font_weight: Constant[FontWeight] = Constant("GRA_FONT_WGHT", FontWeight)
    graphic_font_spacing: Constant[FontSpacing] = Constant(
        "GRA_FONT_SPCING", FontSpacing
    )
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class Item(GenericObject):
    # auto-generated
    object_type = FormsObjects.item

    justification: Constant[Justification] = Constant("JUSTIFICATION", Justification)
    automatic_skip = Bool("AUTO_SKP")
    background_color = Text("BACK_COLOR")
    database_item = Bool("DB_ITM")
    bevel: Constant[Bevel] = Constant("BEVEL", Bevel)
    mirror_item_object: Obj[BaseObject] = Object("SYNC_ITM_OBJ")
    canvas = Text("CNV_NAM")
    canvas_object_pointer: Obj[Canvas] = Object("CNV_OBJ")
    case_info = Unknown("CLIENT_INFO")
    case_insensitive_query = Bool("CASE_INSENSITIVE_QRY")
    case_restriction: Constant[CaseRestriction] = Constant(
        "CASE_RSTRCTION", CaseRestriction
    )
    check_box_mapping_of_other_values: Constant[
        CheckBoxMappingofOtherValues
    ] = Constant("CHK_BX_OTHER_VALS", CheckBoxMappingofOtherValues)
    value_when_checked = Text("CHKED_VAL")
    synchronize_with_item = Text("SYNC_ITM_NAM")
    comments = Text("COMMENT")
    image_format: Constant[ImageFormat] = Constant("IMG_FMT", ImageFormat)
    image_depth: Constant[ImageDepth] = Constant("IMG_DPTH", ImageDepth)
    prompt_foreground_color = Text("PRMPT_FORE_COLOR")
    default_button = Bool("DFLT_BTN")
    initial_keyboard_state: Constant[InitialKeyboardState] = Constant(
        "INIT_KBRD_DIR", InitialKeyboardState
    )
    direction: Constant[Direction] = Constant("LANG_DIR", Direction)
    visible = Bool("VISIBLE")
    editor = Text("EDT_NAM")
    editor_object_pointer: Obj[Editor] = Object("EDT_OBJ")
    editor_x_position = Number("EDT_X_POS")
    editor_y_position = Number("EDT_Y_POS")
    enabled = Bool("ENABLED")
    fill_pattern = Text("FILL_PAT")
    font_name = Text("FONT_NAM")
    font_size = Number("FONT_SIZ")
    font_style: Constant[FontStyle] = Constant("FONT_STY", FontStyle)
    font_weight: Constant[FontWeight] = Constant("FONT_WGHT", FontWeight)
    font_spacing: Constant[FontSpacing] = Constant("FONT_SPCING", FontSpacing)
    foreground_color = Text("FORE_COLOR")
    format_mask = Text("FMT_MSK")
    filename = Text("FLNAM")
    execution_mode: Constant[ExecutionMode] = Constant("EXEC_MODE", ExecutionMode)
    height = Number("HEIGHT")
    highest_allowed_value = Text("HIGHEST_ALLOWED_VAL")
    hint = Text("HINT")
    show_horizontal_scroll_bar = Bool("SHOW_HORZ_SCRLBR")
    iconic = Bool("ICONIC")
    icon_filename = Text("ICON_FLNAM")
    insert_allowed = Bool("INSRT_ALLOWED")
    item_type: Constant[ItemType] = Constant("ITM_TYP", ItemType)
    copy_value_from_item = Text("COPY_VAL_FROM_ITM")
    data_type: Constant[DataType] = Constant("DAT_TYP", DataType)
    initial_value = Text("INIT_VAL")
    keep_cursor_position = Bool("KEEP_CRSR_POS")
    label = Text("LABEL")
    prompt_background_color = Text("PRMPT_BACK_COLOR")
    lock_record = Bool("LOCK_REC")
    list_of_values = Text("LOV_NAM")
    lov_object_pointer: Obj[LOV] = Object("LOV_OBJ")
    validate_from_list = Bool("VALIDATE_FROM_LST")
    list_x_position = Number("LOV_X_POS")
    list_y_position = Number("LOV_Y_POS")
    lowest_allowed_value = Text("LOWEST_ALLOWED_VAL")
    list_style: Constant[ListStyle] = Constant("LST_STY", ListStyle)
    maximum_length = Number("MAX_LEN")
    compression_quality: Constant[CompressionQuality] = Constant(
        "CMPRSSION_QLTY", CompressionQuality
    )
    access_key = Text("ACCESS_KEY")
    mouse_navigate = Bool("MOUSE_NAVIGATE")
    multi_line = Bool("MLT_LIN")
    name = Text("NAME")
    prompt_fill_pattern = Text("PRMPT_FILL_PAT")
    keyboard_navigable = Bool("KBRD_NAVIGABLE")
    next_object: Obj[Item] = Object("NEXT")
    next_navigation_item = Text("NXT_NAVIGATION_ITM_NAM")
    next_item_object: Obj[BaseObject] = Object("NXT_NAVIGATION_ITM_OBJ")
    mapping_of_other_values = Text("OTHER_VALS")
    primary_key = Bool("PRMRY_KEY")
    previous_navigation_item = Text("PREV_NAVIGATION_ITM_NAM")
    previous_item_object: Obj[BaseObject] = Object("PREV_NAVIGATION_ITM_OBJ")
    query_allowed = Bool("QRY_ALLOWED")
    query_length = Number("QRY_LEN")
    query_only = Bool("QRY_ONLY")
    display_quality: Constant[DisplayQuality] = Constant("DISP_QLTY", DisplayQuality)
    radio_button_object: Obj[RadioButton] = Object("RAD_BUT")
    reading_order: Constant[ReadingOrder] = Constant("READING_ORDR", ReadingOrder)
    distance_between_records = Number("DIST_BTWN_RECS")
    record_group = Text("REC_GRP_NAM")
    record_group_object_pointer: Obj[RecordGroup] = Object("REC_GRP_OBJ")
    current_record_visual_attribute_group = Text("REC_VAT_GRP_NAM")
    rendered = Bool("RENDERED")
    required = Bool("REQUIRED")
    prompt_font_name = Text("PRMPT_FONT_NAM")
    conceal_data = Bool("CONCEAL_DATA")
    sizing_style: Constant[SizingStyle] = Constant("SIZING_STY", SizingStyle)
    prompt_font_size = Number("PRMPT_FONT_SIZ")
    triggers: ObjectList[Trigger] = Subobjects("TRIGGER")
    value_when_unchecked = Text("UNCHKED_VAL")
    update_allowed = Bool("UPDT_ALLOWED")
    update_only_if_null = Bool("UPDT_IF_NULL")
    visual_attribute_group = Text("VAT_NAM")
    va_object: Obj[VisualAttribute] = Object("VAT_OBJ")
    prompt_font_style: Constant[FontStyle] = Constant("PRMPT_FONT_STY", FontStyle)
    calculation_mode: Constant[CalculationMode] = Constant("CALC_MODE", CalculationMode)
    prompt_font_weight: Constant[FontWeight] = Constant("PRMPT_FONT_WGHT", FontWeight)
    prompt_font_spacing: Constant[FontSpacing] = Constant(
        "PRMPT_FONT_SPCING", FontSpacing
    )
    column_name = Text("COL_NAM")
    prompt_display_style: Constant[PromptDisplayStyle] = Constant(
        "PRMPT_DISP_STY", PromptDisplayStyle
    )
    show_vertical_scroll_bar = Bool("SHOW_VERT_SCRLBR")
    width = Number("WIDTH")
    wrap_style: Constant[WrapStyle] = Constant("WRAP_STY", WrapStyle)
    x_position = Number("X_POS")
    y_position = Number("Y_POS")
    display_hint_automatically = Bool("AUTO_HINT")
    tooltip = Text("TOOLTIP")
    current_record_va_pointer: Obj[VisualAttribute] = Object("REC_VAT_GRP_OBJ")
    prompt = Text("PRMPT")
    prompt_reading_order: Constant[ReadingOrder] = Constant(
        "PRMPT_READING_ORDR", ReadingOrder
    )
    prompt_justification: Constant[Justification] = Constant("PRMPT_JST", Justification)
    previous_object: Obj[Item] = Object("PREVIOUS")
    prompt_attachment_edge: Constant[PromptAttachmentEdge] = Constant(
        "PRMPT_ATT_EDGE", PromptAttachmentEdge
    )
    prompt_attachment_offset = Number("PRMPT_ATT_OFST")
    prompt_alignment: Constant[PromptAlignment] = Constant(
        "PRMPT_ALIGN", PromptAlignment
    )
    prompt_alignment_offset: Constant[PromptAlignment] = Constant(
        "PRMPT_ALIGN_OFST", PromptAlignment
    )
    owning_object: Obj[BaseObject] = Object("OWNER")
    formula = Text("FORMULA")
    summary_function: Constant[SummaryFunction] = Constant("SUMM_FUNC", SummaryFunction)
    summarized_item = Text("SUMM_ITM_NAM")
    number_of_items_displayed = Number("ITMS_DISP")
    tab_page = Text("TBP_NAM")
    tab_page_name: Obj[BaseObject] = Object("TBP_OBJ")
    communication_mode: Constant[CommunicationMode] = Constant(
        "COMM_MODE", CommunicationMode
    )
    data_source_data_block = Text("DAT_SRC_BLK")
    data_source_x_axis = Text("DAT_SRC_X_AXS")
    data_source_y_axis = Text("DAT_SRC_Y_AXS")
    update_on_commit = Bool("UPDT_COMMIT")
    update_on_query = Bool("UPDT_QRY")
    query_name = Text("QUERY_NAME")
    popup_menu = Text("POPUP_MNU_NAM")
    popup_menu_object: Obj[BaseObject] = Object("POPUP_MNU_OBJ")
    tooltip_visual_attribute_group = Text("TOOLTIP_VAT_GRP")
    tooltip_va_object: Obj[VisualAttribute] = Object("POPUP_VA_OBJ")
    prompt_va_object: Obj[VisualAttribute] = Object("PRMPT_VAT_OBJ")
    prompt_visual_attribute_group = Text("PRMPT_VAT_NAM")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    count_of_list_element_items = Number("LST_ELEMENT_COUNT")
    keyboard_state: Constant[KeyboardState] = Constant("KBRD_STATE", KeyboardState)
    summarized_block = Text("SUMM_BLK_NAM")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_first_level_owners_name = Text("PAR_SL1OBJ_NAM")
    parent_objects_first_level_owners_type = Number("PAR_SL1OBJ_TYP")
    parent_objects_type = Number("PAR_TYP")
    help_book_topic = Text("HELP_BOOK_TOPIC")
    allow_empty_branches = Bool("TRE_ALLW_EMP_BRANCH")
    multi_selection = Bool("TRE_MULTI_SELECT")
    show_lines = Bool("TRE_SHOW_LINES")
    show_symbols = Bool("TRE_SHOW_SYMBOL")
    data_query = Text("TRE_DATA_QRY")
    implementation_class = Text("IMPL_CLASS")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")
    data_length_semantics: Constant[DataLengthSemantics] = Constant(
        "DATA_LEN_SEMANTICS", DataLengthSemantics
    )
    row_banding_frequency = Number("ROW_BANDING_FREQ")
    cursor_style = Number("CURSOR_STYLE")


@forms_object
class Point(GenericObject):
    # auto-generated
    object_type = FormsObjects.point

    name = Text("NAME")
    next_object: Obj[Point] = Object("NEXT")
    x_position = Number("X_POS")
    y_position = Number("Y_POS")
    previous_object: Obj[Point] = Object("PREVIOUS")


@forms_object
class ProgramUnit(GenericObject):
    # auto-generated
    object_type = FormsObjects.program_unit

    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    name = Text("NAME")
    next_object: Obj[ProgramUnit] = Object("NEXT")
    previous_object: Obj[ProgramUnit] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    program_unit_text = Text("PGU_TXT")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    program_unit_type: Constant[ProgramUnitType] = Constant("PGU_TYP", ProgramUnitType)
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class PropertyClass(GenericObject):
    # auto-generated
    object_type = FormsObjects.property_class

    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    name = Text("NAME")
    next_object: Obj[PropertyClass] = Object("NEXT")
    triggers: ObjectList[Trigger] = Subobjects("TRIGGER")
    previous_object: Obj[PropertyClass] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class RadioButton(GenericObject):
    # auto-generated
    object_type = FormsObjects.radio_button

    background_color = Text("BACK_COLOR")
    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    prompt_foreground_color = Text("PRMPT_FORE_COLOR")
    visible = Bool("VISIBLE")
    enabled = Bool("ENABLED")
    fill_pattern = Text("FILL_PAT")
    font_name = Text("FONT_NAM")
    font_size = Number("FONT_SIZ")
    font_style: Constant[FontStyle] = Constant("FONT_STY", FontStyle)
    font_weight: Constant[FontWeight] = Constant("FONT_WGHT", FontWeight)
    font_spacing: Constant[FontSpacing] = Constant("FONT_SPCING", FontSpacing)
    foreground_color = Text("FORE_COLOR")
    height = Number("HEIGHT")
    label = Text("LABEL")
    prompt_background_color = Text("PRMPT_BACK_COLOR")
    access_key = Text("ACCESS_KEY")
    name = Text("NAME")
    prompt_fill_pattern = Text("PRMPT_FILL_PAT")
    next_object: Obj[RadioButton] = Object("NEXT")
    radio_button_value = Text("RDB_VAL")
    distance_between_records = Number("DIST_BTWN_RECS")
    prompt_font_name = Text("PRMPT_FONT_NAM")
    prompt_font_size = Number("PRMPT_FONT_SIZ")
    visual_attribute_group = Text("VAT_NAM")
    va_object: Obj[VisualAttribute] = Object("VAT_OBJ")
    prompt_font_style: Constant[FontStyle] = Constant("PRMPT_FONT_STY", FontStyle)
    prompt_font_weight: Constant[FontWeight] = Constant("PRMPT_FONT_WGHT", FontWeight)
    prompt_font_spacing: Constant[FontSpacing] = Constant(
        "PRMPT_FONT_SPCING", FontSpacing
    )
    prompt_display_style: Constant[PromptDisplayStyle] = Constant(
        "PRMPT_DISP_STY", PromptDisplayStyle
    )
    width = Number("WIDTH")
    x_position = Number("X_POS")
    y_position = Number("Y_POS")
    prompt = Text("PRMPT")
    prompt_reading_order: Constant[ReadingOrder] = Constant(
        "PRMPT_READING_ORDR", ReadingOrder
    )
    prompt_justification: Constant[Justification] = Constant("PRMPT_JST", Justification)
    previous_object: Obj[RadioButton] = Object("PREVIOUS")
    prompt_attachment_edge: Constant[PromptAttachmentEdge] = Constant(
        "PRMPT_ATT_EDGE", PromptAttachmentEdge
    )
    prompt_attachment_offset = Number("PRMPT_ATT_OFST")
    prompt_alignment: Constant[PromptAlignment] = Constant(
        "PRMPT_ALIGN", PromptAlignment
    )
    prompt_alignment_offset: Constant[PromptAlignment] = Constant(
        "PRMPT_ALIGN_OFST", PromptAlignment
    )
    owning_object: Obj[BaseObject] = Object("OWNER")
    prompt_va_object: Obj[VisualAttribute] = Object("PRMPT_VAT_OBJ")
    prompt_visual_attribute_group = Text("PRMPT_VAT_NAM")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_first_level_owners_name = Text("PAR_SL1OBJ_NAM")
    parent_objects_first_level_owners_type = Number("PAR_SL1OBJ_TYP")
    parent_objects_second_level_owners_name = Text("PAR_SL2OBJ_NAM")
    parent_objects_second_level_owners_type = Number("PAR_SL2OBJ_TYP")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class Relation(GenericObject):
    # auto-generated
    object_type = FormsObjects.relation

    automatic_query = Bool("AUTO_QRY")
    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    deferred = Bool("DEFERRED")
    delete_record_behavior: Constant[DeleteRecordBehavior] = Constant(
        "DEL_REC", DeleteRecordBehavior
    )
    detail_data_block = Text("DETAIL_BLK")
    join_condition = Text("JOIN_COND")
    name = Text("NAME")
    next_object: Obj[Relation] = Object("NEXT")
    prevent_masterless_operations = Bool("PRVNT_MSTRLESS_OPS")
    previous_object: Obj[Relation] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    relation_type: Constant[RelationType] = Constant("REL_TYPE", RelationType)
    detail_reference_item = Text("DETAIL_ITEMREF")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class TabPage(GenericObject):
    # auto-generated
    object_type = FormsObjects.tab_page

    background_color = Text("BACK_COLOR")
    graphics: ObjectList[Graphic] = Subobjects("GRAPHIC")
    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    visible = Bool("VISIBLE")
    enabled = Bool("ENABLED")
    fill_pattern = Text("FILL_PAT")
    foreground_color = Text("FORE_COLOR")
    icon_filename = Text("ICON_FLNAM")
    label = Text("LABEL")
    name = Text("NAME")
    next_object: Obj[TabPage] = Object("NEXT")
    visual_attribute_group = Text("VAT_NAM")
    previous_object: Obj[TabPage] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_first_level_owners_name = Text("PAR_SL1OBJ_NAM")
    parent_objects_first_level_owners_type = Number("PAR_SL1OBJ_TYP")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")
    gradient_start_side = Number("GRADIENT_START")


@forms_object
class Trigger(GenericObject):
    # auto-generated
    object_type = FormsObjects.trigger

    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    execution_hierarchy: Constant[ExecutionHierarchy] = Constant(
        "EXEC_HIERARCHY", ExecutionHierarchy
    )
    fire_in_enter_query_mode = Bool("FIRE_IN_QRY")
    name = Text("NAME")
    next_object: Obj[Trigger] = Object("NEXT")
    hide_object = Bool("HIDE")
    display_in_keyboard_help = Bool("DISP_IN_KBRD_HLP")
    keyboard_help_text = Text("KBRD_HLP_TXT")
    trigger_step_object: Obj[BaseObject] = Object("TRIG_STEP")
    trigger_style: Constant[TriggerStyle] = Constant("TRG_STY", TriggerStyle)
    trigger_text = Text("TRG_TXT")
    previous_object: Obj[Trigger] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    trigger_internal_type = Number("TRG_INTERNAL_TYP")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_first_level_owners_name = Text("PAR_SL1OBJ_NAM")
    parent_objects_first_level_owners_type = Number("PAR_SL1OBJ_TYP")
    parent_objects_second_level_owners_name = Text("PAR_SL2OBJ_NAM")
    parent_objects_second_level_owners_type = Number("PAR_SL2OBJ_TYP")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


# todo: todo: undocumented, no .h file
@forms_object
class VisualAttribute(GenericObject):
    # auto-generated
    object_type = FormsObjects.visual_attribute

    background_color = Text("BACK_COLOR")
    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    prompt_foreground_color = Text("PRMPT_FORE_COLOR")
    fill_pattern = Text("FILL_PAT")
    font_name = Text("FONT_NAM")
    font_size = Number("FONT_SIZ")
    font_style: Constant[FontStyle] = Constant("FONT_STY", FontStyle)
    font_weight: Constant[FontWeight] = Constant("FONT_WGHT", FontWeight)
    font_spacing: Constant[FontSpacing] = Constant("FONT_SPCING", FontSpacing)
    foreground_color = Text("FORE_COLOR")
    prompt_background_color = Text("PRMPT_BACK_COLOR")
    name = Text("NAME")
    prompt_fill_pattern = Text("PRMPT_FILL_PAT")
    next_object: Obj[VisualAttribute] = Object("NEXT")
    prompt_font_name = Text("PRMPT_FONT_NAM")
    prompt_font_size = Number("PRMPT_FONT_SIZ")
    prompt_font_style: Constant[FontStyle] = Constant("PRMPT_FONT_STY", FontStyle)
    prompt_font_weight: Constant[FontWeight] = Constant("PRMPT_FONT_WGHT", FontWeight)
    prompt_font_spacing: Constant[FontSpacing] = Constant(
        "PRMPT_FONT_SPCING", FontSpacing
    )
    previous_object: Obj[VisualAttribute] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    visual_attribute_type: Constant[VisualAttributeType] = Constant(
        "VAT_TYP", VisualAttributeType
    )
    frame_title_foreground_color = Text("FRAME_TTL_FORE_COLOR")
    frame_title_background_color = Text("FRAME_TTL_BACK_COLOR")
    frame_title_fill_pattern = Text("FRAME_TTL_FILL_PAT")
    frame_title_font_name = Text("FRAME_TTL_FONT_NAM")
    frame_title_font_size = Number("FRAME_TTL_FONT_SIZ")
    frame_title_font_style: Constant[FontStyle] = Constant(
        "FRAME_TTL_FONT_STY", FontStyle
    )
    frame_title_font_weight: Constant[FontWeight] = Constant(
        "FRAME_TTL_FONT_WGHT", FontWeight
    )
    frame_title_font_spacing: Constant[FontSpacing] = Constant(
        "FRAME_TTL_FONT_SPCING", FontSpacing
    )
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class VisualState(GenericObject):
    # auto-generated
    object_type = FormsObjects.visual_state


@forms_object
class Window(GenericObject):
    # auto-generated
    object_type = FormsObjects.window

    background_color = Text("BACK_COLOR")
    bevel: Constant[Bevel] = Constant("BEVEL", Bevel)
    case_info = Unknown("CLIENT_INFO")
    close_allowed = Bool("CLS_ALLOWED")
    comments = Text("COMMENT")
    direction: Constant[Direction] = Constant("LANG_DIR", Direction)
    fill_pattern = Text("FILL_PAT")
    resize_allowed = Bool("RESIZE_ALLOWED")
    font_name = Text("FONT_NAM")
    font_size = Number("FONT_SIZ")
    font_style: Constant[FontStyle] = Constant("FONT_STY", FontStyle)
    font_weight: Constant[FontWeight] = Constant("FONT_WGHT", FontWeight)
    font_spacing: Constant[FontSpacing] = Constant("FONT_SPCING", FontSpacing)
    foreground_color = Text("FORE_COLOR")
    height = Number("HEIGHT")
    show_horizontal_scroll_bar = Bool("SHOW_HORZ_SCRLBR")
    horizontal_toolbar_canvas = Text("HTB_CNV_NAME")
    minimize_allowed = Bool("MINIMIZE_ALLOWED")
    icon_filename = Text("ICON_FLNAM")
    minimized_title = Text("MINIMIZE_TTL")
    inherit_menu = Bool("INHRT_MNU")
    modal = Bool("MODAL")
    move_allowed = Bool("MV_ALLOWED")
    name = Text("NAME")
    next_object: Obj[Window] = Object("NEXT")
    hide_on_exit = Bool("HIDE_ON_EXIT")
    title = Text("TITLE")
    visual_attribute_group = Text("VAT_NAM")
    va_object: Obj[VisualAttribute] = Object("VAT_OBJ")
    show_vertical_scroll_bar = Bool("SHOW_VERT_SCRLBR")
    vertical_toolbar_canvas = Text("VTB_CNV_NAME")
    width = Number("WIDTH")
    window_style: Constant[WindowStyle] = Constant("WIN_STY", WindowStyle)
    primary_canvas = Text("PRMRY_CNV")
    x_position = Number("X_POS")
    y_position = Number("Y_POS")
    maximize_allowed = Bool("MAXIMIZE_ALLOWED")
    previous_object: Obj[Window] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_type = Number("PAR_TYP")
    help_book_topic = Text("HELP_BOOK_TOPIC")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


# todo: this is special, has own create functions in C
@forms_object
class DataSourceArgument(GenericObject):
    # auto-generated
    object_type = FormsObjects.data_source_argument

    case_info = Unknown("CLIENT_INFO")
    next_object: Obj[DataSourceArgument] = Object("NEXT")
    previous_object: Obj[DataSourceArgument] = Object("PREVIOUS")
    argument_mode: Constant[ArgumentMode] = Constant("DSA_MODE", ArgumentMode)
    argument_name = Text("DSA_NAM")
    argument_type: Constant[ArgumentType] = Constant("DSA_TYP", ArgumentType)
    argument_value = Text("DSA_VAL")
    argument_type_name = Text("DSA_TYP_NAM")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


# todo: this is special, has own create functions in C
@forms_object
class DataSourceColumn(GenericObject):
    # auto-generated
    object_type = FormsObjects.data_source_column

    case_info = Unknown("CLIENT_INFO")
    next_object: Obj[DataSourceColumn] = Object("NEXT")
    previous_object: Obj[DataSourceColumn] = Object("PREVIOUS")
    data_source_column_name = Text("DSC_NAM")
    data_source_column_type: Constant[ColumnType] = Constant("DSC_TYP", ColumnType)
    length = Number("DSC_LEN")
    mandatory = Bool("DSC_MANDATORY")
    precision = Number("DSC_PRECISION")
    scale = Number("DSC_SCALE")
    column_type_name = Text("DSC_TYPE_NAME")
    parent_column = Text("DSC_PARENT_NAME")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class Editor(GenericObject):
    # auto-generated
    object_type = FormsObjects.editor

    background_color = Text("BACK_COLOR")
    bottom_title = Text("BTM_TTL")
    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    fill_pattern = Text("FILL_PAT")
    font_name = Text("FONT_NAM")
    font_size = Number("FONT_SIZ")
    font_style: Constant[FontStyle] = Constant("FONT_STY", FontStyle)
    font_weight: Constant[FontWeight] = Constant("FONT_WGHT", FontWeight)
    font_spacing: Constant[FontSpacing] = Constant("FONT_SPCING", FontSpacing)
    foreground_color = Text("FORE_COLOR")
    height = Number("HEIGHT")
    show_horizontal_scroll_bar = Bool("SHOW_HORZ_SCRLBR")
    name = Text("NAME")
    next_object: Obj[Editor] = Object("NEXT")
    title = Text("TITLE")
    visual_attribute_group = Text("VAT_NAM")
    va_object: Obj[VisualAttribute] = Object("VAT_OBJ")
    show_vertical_scroll_bar = Bool("SHOW_VERT_SCRLBR")
    width = Number("WIDTH")
    wrap_style: Constant[WrapStyle] = Constant("WRAP_STY", WrapStyle)
    x_position = Number("X_POS")
    y_position = Number("Y_POS")
    previous_object: Obj[Editor] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class LOV(GenericObject):
    # auto-generated
    object_type = FormsObjects.lov

    automatic_select = Bool("AUTO_SLCT")
    automatic_display = Bool("AUTO_DISP")
    automatic_refresh = Bool("AUTO_RFRSH")
    automatic_skip = Bool("AUTO_SKP")
    background_color = Text("BACK_COLOR")
    case_info = Unknown("CLIENT_INFO")
    column_mapping_object: Obj[LOVColumnMap] = Object("COL_MAP")
    comments = Text("COMMENT")
    direction: Constant[Direction] = Constant("LANG_DIR", Direction)
    fill_pattern = Text("FILL_PAT")
    font_name = Text("FONT_NAM")
    font_size = Number("FONT_SIZ")
    font_style: Constant[FontStyle] = Constant("FONT_STY", FontStyle)
    font_weight: Constant[FontWeight] = Constant("FONT_WGHT", FontWeight)
    font_spacing: Constant[FontSpacing] = Constant("FONT_SPCING", FontSpacing)
    foreground_color = Text("FORE_COLOR")
    height = Number("HEIGHT")
    filter_before_display = Bool("FLTR_BEFORE_DISP")
    list_type: Constant[ListType] = Constant("LST_TYP", ListType)
    name = Text("NAME")
    next_object: Obj[LOV] = Object("NEXT")
    old_lov_text = Text("OLD_LOV_TXT")
    record_group = Text("REC_GRP_NAM")
    record_group_object_pointer: Obj[RecordGroup] = Object("REC_GRP_OBJ")
    title = Text("TITLE")
    visual_attribute_group = Text("VAT_NAM")
    va_object: Obj[VisualAttribute] = Object("VAT_OBJ")
    width = Number("WIDTH")
    x_position = Number("X_POS")
    y_position = Number("Y_POS")
    previous_object: Obj[LOV] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    automatic_position = Bool("AUTO_POS")
    automatic_column_width = Bool("AUTO_COL_WID")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class LOVColumnMap(GenericObject):
    # auto-generated
    object_type = FormsObjects.lov_column_map

    case_info = Unknown("CLIENT_INFO")
    display_width = Number("DISP_WID")
    name = Text("NAME")
    next_object: Obj[LOVColumnMap] = Object("NEXT")
    return_item = Text("RTRN_ITM")
    title = Text("TITLE")
    previous_object: Obj[LOVColumnMap] = Object("PREVIOUS")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class Menu(GenericObject):
    # auto-generated
    object_type = FormsObjects.menu

    bottom_title = Text("BTM_TTL")
    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    menu_item_object: Obj[MenuItem] = Object("MNU_ITM")
    name = Text("NAME")
    next_object: Obj[Menu] = Object("NEXT")
    sub_title = Text("SUB_TTL")
    tear_off_menu = Bool("TEAR_OFF_MNU")
    title = Text("TITLE")
    previous_object: Obj[Menu] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class MenuItem(GenericObject):
    # auto-generated
    object_type = FormsObjects.menu_item

    keyboard_accelerator = Text("KBRD_ACC")
    case_info = Unknown("CLIENT_INFO")
    command_text = Text("COM_TXT")
    command_type: Constant[CommandType] = Constant("COM_TYP", CommandType)
    comments = Text("COMMENT")
    visible = Bool("VISIBLE")
    display_without_privilege = Bool("DISP_NO_PRIV")
    enabled = Bool("ENABLED")
    font_name = Text("FONT_NAM")
    font_size = Number("FONT_SIZ")
    font_style: Constant[FontStyle] = Constant("FONT_STY", FontStyle)
    font_weight: Constant[FontWeight] = Constant("FONT_WGHT", FontWeight)
    font_spacing: Constant[FontSpacing] = Constant("FONT_SPCING", FontSpacing)
    hint = Text("HINT")
    icon_filename = Text("ICON_FLNAM")
    label = Text("LABEL")
    magic_item: Constant[MagicItem] = Constant("MAGIC_ITM", MagicItem)
    menu_item_type: Constant[ItemType] = Constant("MNU_ITM_TYP", ItemType)
    menu_item_radio_group = Text("MNU_ITM_RAD_GRP")
    name = Text("NAME")
    next_object: Obj[MenuItem] = Object("NEXT")
    visual_attribute_group = Text("VAT_NAM")
    va_object: Obj[VisualAttribute] = Object("VAT_OBJ")
    previous_object: Obj[MenuItem] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    visible_in_vertical_menu_toolbar = Bool("VSBL_IN_VERT_MNU_TLBR")
    visible_in_horizontal_menu_toolbar = Bool("VSBL_IN_HORZ_MNU_TLBR")
    visible_in_menu = Bool("VSBL_IN_MENU")
    icon_in_menu = Bool("ICON_IN_MNU")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    count_of_roles = Number("ROLE_COUNT")
    submenu_name = Text("SUB_MNU_NAM")
    submenu_object: Obj[BaseObject] = Object("SUB_MNU_OBJ")
    menu_item_code = Text("MNU_ITM_CODE")
    parent_objects_first_level_owners_name = Text("PAR_SL1OBJ_NAM")
    parent_objects_first_level_owners_type = Number("PAR_SL1OBJ_TYP")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class ObjectGroup(GenericObject):
    # auto-generated
    object_type = FormsObjects.object_group

    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    name = Text("NAME")
    next_object: Obj[ObjectGroup] = Object("NEXT")
    object_group_child_object: Obj[ObjectGroup] = Object("OG_CHILD")
    previous_object: Obj[ObjectGroup] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    object_group_type = Number("OBJ_GRP_TYP")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class ObjectChild(GenericObject):
    # auto-generated
    object_type = FormsObjects.object_child

    case_info = Unknown("CLIENT_INFO")
    name = Text("NAME")
    next_object: Obj[ObjectChild] = Object("NEXT")
    previous_object: Obj[ObjectChild] = Object("PREVIOUS")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    real_object_pointed_to_by_the_object_group_child: Obj[BaseObject] = Object(
        "OBJ_GRP_CHILD_REAL_OBJ"
    )
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class RecordGroup(GenericObject):
    # auto-generated
    object_type = FormsObjects.record_group

    case_info = Unknown("CLIENT_INFO")
    record_group_colspecs: ObjectList[RecordGroupColspec] = Subobjects("COL_SPEC")
    comments = Text("COMMENT")
    name = Text("NAME")
    next_object: Obj[RecordGroup] = Object("NEXT")
    record_group_query = Text("REC_GRP_QRY")
    record_group_type: Constant[RecordGroupType] = Constant(
        "REC_GRP_TYP", RecordGroupType
    )
    previous_object: Obj[RecordGroup] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    record_group_fetch_size = Number("REC_GRP_FETCH_SIZ")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")


@forms_object
class RecordGroupColspec(GenericObject):
    # auto-generated
    object_type = FormsObjects.record_group_colspec

    case_info = Unknown("CLIENT_INFO")
    column_data_type: Constant[ColumnDataType] = Constant("COL_DAT_TYP", ColumnDataType)
    maximum_length = Number("MAX_LEN")
    name = Text("NAME")
    next_object: Obj[RecordGroupColspec] = Object("NEXT")
    previous_object: Obj[RecordGroupColspec] = Object("PREVIOUS")
    count_of_data_items_in_the_column = Number("COL_VALS_COUNT")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")
    column_value: ObjectList[ColumnValue] = Subobjects("COLUMN_VALUE")
    data_length_semantics: Constant[DataLengthSemantics] = Constant(
        "DATA_LEN_SEMANTICS", DataLengthSemantics
    )


@forms_object
class Report(GenericObject):
    # auto-generated
    object_type = FormsObjects.report

    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    filename = Text("FLNAM")
    execution_mode: Constant[ExecutionMode] = Constant("EXEC_MODE", ExecutionMode)
    name = Text("NAME")
    next_object: Obj[Report] = Object("NEXT")
    previous_object: Obj[Report] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    communication_mode: Constant[CommunicationMode] = Constant(
        "COMM_MODE", CommunicationMode
    )
    data_source_data_block = Text("DAT_SRC_BLK")
    query_name = Text("QUERY_NAME")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    report_destination_type: Constant[ReportDestinationType] = Constant(
        "RPT_DESTINATION_TYP", ReportDestinationType
    )
    report_destination_name = Text("RPT_DESTINATION_NAM")
    report_destination_format = Text("RPT_DESTINATION_FMT")
    report_server = Text("RPT_SRVR")
    other_reports_parameters = Text("RPT_PARAMS")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")
    report_object_type = Number("RPT_OBJECT_TYPE")
    delivery_type = Number("RPT_DEL_TYPE")
    ssl_connection = Number("RPT_SSL_CONN")
    service_location = Text("RPT_SRVC_LOC")
    report_path = Text("RPT_ABS_PATH")
    report_locale = Text("RPT_LOCALE")
    report_format = Text("RPT_OPT_FMT")
    template_name = Text("RPT_TEMPLATE_NAME")
    report_parameters = Text("RPT_BIP_PARAMS")
    ftp_server_name = Text("RPT_FTP_SERVER")
    ftp_user = Text("RPT_FTP_USER")
    ftp_file_name = Text("RPT_FTP_FILENAME")
    secure_ftp = Bool("RPT_FTP_SECURED")
    number = Text("RPT_FAX_NUMBER")
    fax_server_name = Text("RPT_FAX_SERVER")
    to_email = Text("RPT_MAIL_TO")
    from_email = Text("RPT_MAIL_FROM")
    cc_email = Text("RPT_MAIL_CC")
    bcc_email = Text("RPT_MAIL_BCC")
    reply_to = Text("RPT_MAIL_REPLYTO")
    subject = Text("RPT_MAIL_SUBJECT")
    body = Text("RPT_MAIL_BODY")
    mail_server_name = Text("RPT_MAIL_SERVER")
    local_file_name = Text("RPT_LOCAL_FILENAME")
    number_of_copies = Number("RPT_PRINT_NUMBEROFCOPY")
    orientation = Number("RPT_PRINT_ORIENTATION")
    page_range = Text("RPT_PRINT_PAGERANGE")
    side = Number("RPT_PRINT_SIDE")
    tray = Number("RPT_PRINT_TRAY")
    print_server_name = Text("RPT_PRINT_NAME")
    webdav_auth_type = Number("RPT_WEBDAV_AUTHTYPE")
    webdav_user = Text("RPT_WEBDAV_USER")
    webdav_file_path = Text("RPT_WEBDAV_FILE")
    webdav_server_name = Text("RPT_WEBDAV_SERVER")


@forms_object
class Event(GenericObject):
    # auto-generated
    object_type = FormsObjects.event

    case_info = Unknown("CLIENT_INFO")
    comments = Text("COMMENT")
    name = Text("NAME")
    next_object: Obj[Event] = Object("NEXT")
    triggers: ObjectList[Trigger] = Subobjects("TRIGGER")
    previous_object: Obj[Event] = Object("PREVIOUS")
    owning_object: Obj[BaseObject] = Object("OWNER")
    owning_module: Obj[Module] = Object("MODULE")
    source_object: Obj[PropertyClass] = Object("SOURCE")
    parent_objects_module = Text("PAR_MODULE")
    parent_objects_module_type = Number("PAR_MODTYP")
    parent_object_name = Text("PAR_NAM")
    parent_objects_file_name = Text("PAR_FLNAM")
    parent_objects_file_path = Text("PAR_FLPATH")
    parent_objects_type = Number("PAR_TYP")
    persistent_client_info_storage = Unknown("PERSIST_CLIENT_INFO")
    persistent_client_info_storage_length = Number("PERSIST_CLT_INF_LEN")
    event_type: Constant[EventType] = Constant("EVENT_TYPE", EventType)
    subscription_name = Text("EVENT_SUBS_NAME")
    scope: Constant[Scope] = Constant("EVENT_SCOPE", Scope)
    auto_subscribe = Bool("EVENT_ENABLED")
    event_implementation_class = Text("EVENT_IMPLCLASS")
    correlation_id = Text("EVENT_CORRID")
    priority_mode = Number("EVENT_PRIORITY_MODE")
    view_mode: Constant[ViewMode] = Constant("EVENT_VIEW_MODE", ViewMode)


@forms_object
class ColumnValue(GenericObject):
    # auto-generated
    object_type = FormsObjects.column_value



from __future__ import annotations

import atexit
from ctypes import pointer, c_int, c_void_p, c_char_p, c_bool, c_uint, CDLL
from functools import partial
from typing import (
    Callable,
    Any,
    Optional,
    Tuple,
    TYPE_CHECKING,
    cast,
    Union,
    Type,
    TypeVar,
)

from .error_handling import raise_for_code
from .forms_api import dlls

if TYPE_CHECKING:  # pragma: no cover
    from ctypes import _FuncPointer
    from .generic_object import BaseObject
    from .forms_objects import Module

T = TypeVar("T")

Setter = Callable[["BaseObject", int, T], None]
Getter = Callable[["BaseObject", int], T]
CTypes = Union[Type[c_void_p], Type[c_bool], Type[c_int], Type["String"]]


class Context:
    version: str
    encoding: str
    api: Optional[CDLL]
    free: Optional[_FuncPointer]

    def __init__(self) -> None:
        self.version, self.encoding = "12c", "utf-8"
        self.api, self.free = None, None
        self._as_parameter_ = c_void_p(0)

    def __bool__(self) -> bool:
        return bool(self._as_parameter_)

    def init(self, version: str, encoding: str) -> None:
        if not self:
            self.version, self.encoding = version, encoding
            self.api, msvcrt = dlls(self.version)
            self.free = msvcrt.free
            self.create_context()
            atexit.register(self.destroy_context)

    def create_context(self) -> None:
        # todo: maybe better way than just an assert?
        assert self.api is not None

        ctx = c_void_p()

        func = self.api.d2fctxcr_Create
        func.argtypes = (c_void_p, c_void_p)
        error_code = func(pointer(ctx), pointer(c_int()))
        if error_code:  # pragma: no cover
            raise_for_code(error_code)

        self._as_parameter_ = ctx

    def destroy_context(self) -> None:
        if self._as_parameter_:
            handled_api_function("d2fctxde_Destroy", tuple())(self)
            self._as_parameter_ = c_void_p(0)


context: Context = Context()


class String(c_char_p):
    def __init__(self) -> None:
        # todo: maybe better way than just an assert?
        assert context.free is not None
        super().__init__()
        self.free = context.free

    def __del__(self) -> None:
        self.free(self)


def api_function(  # type: ignore
    api_function_name: str, arguments: Tuple[Any, ...]
) -> Callable[..., int]:
    api_func = getattr(context.api, api_function_name)
    api_func.argtypes = (c_void_p,) + arguments
    return partial(api_func, context)


def inject_return_value(  # type: ignore
    args: Tuple[Any, ...], return_value_index: Optional[int]
) -> Tuple[Any, ...]:
    if return_value_index is not None:
        func_args = list(args)
        return_value = func_args[return_value_index]
        func_args[return_value_index] = pointer(return_value)
        injected_args = tuple(func_args)
    else:
        injected_args, return_value = args, None
    return injected_args, return_value


def handled_api_function(  # type: ignore
    api_function_name: str,
    arguments: Tuple[Any, ...],
    return_value_index: Optional[int] = None,
) -> Callable[..., Any]:
    def _handled_api_function(*args: Any) -> Any:  # type: ignore
        injected_args, return_value = inject_return_value(args, return_value_index)

        error_code = api_function(api_function_name, arguments)(*injected_args)

        if error_code:
            raise_for_code(error_code)
        return return_value

    return _handled_api_function


def handle_return_value(result: int) -> bool:
    if result in (2, 3):  # YES, NO
        return bool(result == 2)
    raise_for_code(result)


def is_subclassed(generic_object: BaseObject) -> bool:
    func = api_function("d2fobis_IsSubclassed", (c_void_p,))
    result = func(generic_object)
    return handle_return_value(result)


def has_property(generic_object: BaseObject, property_number: int) -> bool:
    func = api_function("d2fobhp_HasProp", (c_void_p, c_int))
    result = func(generic_object, property_number)
    return handle_return_value(result)


def setter(function_name: str, setter_type: CTypes) -> Setter[T]:
    return handled_api_function(function_name, (c_void_p, c_int, setter_type))


set_text: Setter[bytes] = setter("d2fobst_SetTextProp", c_void_p)
set_boolean: Setter[bool] = setter("d2fobsb_SetBoolProp", c_bool)
set_number: Setter[int] = setter("d2fobsn_SetNumProp", c_int)
set_object: Setter[BaseObject] = setter("d2fobso_SetObjProp", c_void_p)


def getter(function_name: str, return_type: CTypes) -> Getter[T]:
    func = handled_api_function(function_name, (c_void_p, c_int, c_void_p), 2)

    def _getter(generic_object: BaseObject, property_number: int) -> Getter[T]:
        return func(generic_object, property_number, return_type()).value

    return cast(Getter[T], _getter)


get_boolean: Getter[bool] = getter("d2fobgb_GetBoolProp", c_bool)
get_number: Getter[int] = getter("d2fobgn_GetNumProp", c_int)
get_object: Getter[BaseObject] = getter("d2fobgo_GetObjProp", c_void_p)
get_text: Getter[bytes] = getter("d2fobgt_GetTextProp", String)


def load_module(form_path: str) -> c_void_p:
    return handled_api_function(
        "d2ffmdld_Load", (c_void_p, c_char_p, c_bool), return_value_index=0
    )(c_void_p(), form_path.encode(context.encoding), False)


def create_module(name: str) -> c_void_p:
    return handled_api_function(
        "d2ffmdcr_Create", (c_void_p, c_char_p), return_value_index=0
    )(c_void_p(), name.encode(context.encoding))


def save_module(module: Module, path: str) -> None:
    handled_api_function("d2ffmdsv_Save", (c_void_p, c_char_p, c_bool))(
        module, path.encode(context.encoding), False
    )


def create(owner: BaseObject, name: str, obj_number: int) -> c_void_p:
    return handled_api_function(
        "d2fobcr_Create", (c_void_p, c_void_p, c_char_p, c_int), return_value_index=1
    )(owner, c_void_p(), name.encode(context.encoding), obj_number)


def destroy(generic_object: BaseObject) -> None:
    handled_api_function("d2fobde_Destroy", (c_void_p,))(generic_object)


def move(generic_object: BaseObject, next_object: Optional[BaseObject]) -> None:
    handled_api_function("d2fobmv_Move", (c_void_p, c_void_p))(
        generic_object, next_object
    )


def query_type(generic_object: Union[BaseObject, c_void_p]) -> int:
    return int(
        handled_api_function("d2fobqt_QueryType", (c_void_p,), return_value_index=1)(
            generic_object, c_int()
        ).value
    )


GetConstant = Callable[[int], str]


def get_constant(function_name: str) -> GetConstant:
    def _get_constant(constant_property: int) -> str:
        constant_value = handled_api_function(
            function_name, (c_int, c_void_p), return_value_index=1
        )(constant_property, c_char_p())
        return (constant_value.value or b"").decode(context.encoding)

    return _get_constant


object_name: GetConstant = get_constant("d2fobgcn_GetConstName")
property_constant_name: GetConstant = get_constant("d2fprgcn_GetConstName")
property_name: GetConstant = get_constant("d2fprgn_GetName")


def property_type(property_number: int) -> int:
    return int(api_function("d2fprgt_GetType", (c_uint,))(property_number))


def property_constant_number(property_const_name: str) -> int:
    return handled_api_function(
        "d2fprgcv_GetConstValue", (c_char_p, c_void_p), return_value_index=1
    )(property_const_name.encode(context.encoding), c_int()).value


def object_number(obj_name: str) -> int:
    return int(
        handled_api_function(
            "d2fobgcv_GetConstValue", (c_char_p, c_void_p), return_value_index=1
        )(obj_name.encode(context.encoding), c_int()).value
    )


def set_subclass(
    to_subclass: BaseObject, parent: BaseObject, keep_path: bool = False
) -> None:
    handled_api_function("d2fobsc_SubClass", (c_void_p, c_void_p, c_bool))(
        to_subclass, parent, keep_path
    )


def remove_subclass(to_un_subclass: BaseObject) -> None:
    handled_api_function("d2fobus_UnSubClass", (c_void_p,))(to_un_subclass)



from __future__ import annotations

import enum
from typing import (
    Dict,
    Type,
    Tuple,
    Union,
    NoReturn,
    Iterable,
    List,
    TypeVar,
    Generic,
    Any,
    TYPE_CHECKING,
    Optional,
)

from .context import context, property_type
from .context import get_boolean
from .context import get_number
from .context import get_object
from .context import get_text
from .context import object_name
from .context import property_constant_name
from .context import property_constant_number
from .context import property_name
from .context import query_type
from .context import set_boolean
from .context import set_number
from .context import set_object
from .context import set_text
from .generic_object import BaseObject, ValueTypes, GenericObject
from .property_types import Properties
from .constants import Justification

if TYPE_CHECKING:  # pragma: no cover
    from . import Module

registered_objects: Dict[str, Union[Type[GenericObject], Type["Module"]]] = {}


class ObjectProperties(enum.Enum):
    canvases = "D2FP_CANVAS"
    alerts = "D2FP_ALERT"
    attached_libraries = "D2FP_ATT_LIB"
    data_blocks = "D2FP_BLOCK"
    form_parameters = "D2FP_FORM_PARAM"
    graphics = "D2FP_GRAPHIC"
    items = "D2FP_ITEM"
    points = "D2FP_POINT"
    program_units = "D2FP_PROG_UNIT"
    property_classes = "D2FP_PROP_CLASS"
    radio_buttons = "D2FP_RADIO_BUTTON"
    relations = "D2FP_RELATION"
    tab_pages = "D2FP_TAB_PAGE"
    triggers = "D2FP_TRIGGER"
    visual_attributes = "D2FP_VIS_ATTR"
    windows = "D2FP_WINDOW"
    data_source_arguments = "D2FP_DAT_SRC_ARG"
    data_source_columns = "D2FP_DAT_SRC_COL"
    editors = "D2FP_EDITOR"
    event = "D2FP_EVENT"
    lovs = "D2FP_LOV"
    lov_column_maps = "D2FP_LV_COLMAP"
    menus = "D2FP_MENU"
    menu_items = "D2FP_MENU_ITEM"
    object_groups = "D2FP_OBJ_GRP"
    object_children = "D2FP_OBG_CHILD"
    record_groups = "D2FP_REC_GRP"
    record_group_colspecs = "D2FP_COL_SPEC"
    reports = "D2FP_REPORT"
    column_value = "D2FP_COLUMN_VALUE"


class Common:
    def __init__(self, constant: str):
        self.constant = constant


class Unknown(Common):
    def __get__(self, instance: BaseObject, owner: Type[BaseObject]) -> None:
        return None

    def __set__(self, instance: BaseObject, value: None) -> NoReturn:
        raise NotImplementedError()


class Text(Common):
    def __get__(self, instance: BaseObject, owner: Type[BaseObject]) -> str:
        return (
            get_text(instance, property_constant_number(self.constant)) or b""
        ).decode(context.encoding)

    def __set__(self, instance: BaseObject, value: str) -> None:
        set_text(
            instance,
            property_constant_number(self.constant),
            value.encode(context.encoding),
        )


U = TypeVar("U")


class BasicAttribute(Common, Generic[U]):
    @staticmethod
    def _getter(instance: BaseObject, property_constant: int) -> U:
        raise NotImplementedError()  # pragma: no cover

    @staticmethod
    def _setter(instance: BaseObject, property_constant: int, value: U) -> U:
        raise NotImplementedError()  # pragma: no cover

    def __get__(self, instance: BaseObject, owner: Type[BaseObject]) -> U:
        return self._getter(instance, property_constant_number(self.constant))

    def __set__(self, instance: BaseObject, value: U) -> None:
        self._setter(instance, property_constant_number(self.constant), value)


class Bool(BasicAttribute[bool]):
    _getter = staticmethod(get_boolean)  # type: ignore
    _setter = staticmethod(set_boolean)  # type: ignore


class Number(BasicAttribute[int]):
    _getter = staticmethod(get_number)  # type: ignore
    _setter = staticmethod(set_number)  # type: ignore


class Constant(Common, Generic[U]):
    _getter = staticmethod(get_number)
    _setter = staticmethod(set_number)

    def __init__(self, constant: str, klass: Type[U]):
        self.constant, self.klass = constant, klass

    def __get__(self, instance: BaseObject, owner: Type[BaseObject]) -> U:
        return self.klass(int(self._getter(instance, property_constant_number(self.constant))))  # type: ignore

    def __set__(self, instance: BaseObject, value: U) -> None:
        to_set = (
            value.value if isinstance(value, self.klass) else self.klass(value).value  # type: ignore
        )
        self._setter(instance, property_constant_number(self.constant), int(to_set))


T = TypeVar("T")


class Object(Common, Generic[T]):
    def __get__(
        self, instance: BaseObject, owner: Type[BaseObject]
    ) -> Optional[BaseObject]:
        obj = get_object(instance, property_constant_number(self.constant))
        if obj:
            klass = get_object_constructor(obj)
            return klass(obj)
        return None

    def __set__(self, instance: BaseObject, value: BaseObject) -> None:
        set_object(instance, property_constant_number(self.constant), value)


properties = {
    ValueTypes.UNKNOWN: Unknown,
    ValueTypes.BOOLEAN: Bool,
    ValueTypes.NUMBER: Number,
    ValueTypes.TEXT: Text,
    ValueTypes.OBJECT: Object,
}


def get_object_constructor(obj: BaseObject) -> Union[Type[Module], Type[GenericObject]]:
    obj_name = object_name(query_type(obj))
    klass = registered_objects.get(obj_name, GenericObject)
    return klass


class Subobjects(Generic[T]):
    def __init__(self, constant: str) -> None:
        self.constant = constant

    # todo: this generates a new list every time would be better if its the same list, then
    #  could operate on that list object creation/deletion and you won't have multiple inconsistent lists
    def __get__(
        self, instance: GenericObject, owner: Type[GenericObject]
    ) -> List[BaseObject]:
        def gen_subobjects() -> Iterable[BaseObject]:
            first_child = get_object(instance, property_constant_number(self.constant))
            if first_child:
                klass = get_object_constructor(first_child)

                child = klass(first_child)
                while child:
                    yield child
                    child = klass(child.next_object)  # type: ignore

        subobjects = list(gen_subobjects())
        return subobjects

    def __set__(self, instance: BaseObject, value: List[BaseObject]) -> NoReturn:
        raise AttributeError("can't set attribute")


# todo: kinda redundant since i have property_types.Properties
def property_attribute(
    property_number: int,
) -> Tuple[str, Union[Common, Subobjects[BaseObject]]]:
    constant_name = property_constant_name(property_number)
    const_name = f"D2FP_{constant_name}"
    try:
        obj_property = ObjectProperties(const_name)
    except ValueError:
        prop_name = (
            "_".join(property_name(property_number).lower().split())
            .replace("'", "")
            .replace("-", "_")
            .replace("/", "_")
        )
        value_type = ValueTypes(property_type(property_number=property_number))
        klass = properties[value_type]
        return prop_name, klass(constant_name)
    else:
        prop_name = obj_property.name
        return prop_name, Subobjects(property_constant_name(property_number))


def object_type(cls: Type[BaseObject], api_objects: Dict) -> Tuple[Dict, int]:  # type: ignore
    try:
        obj_type = api_objects[cls.object_type.value]
    except KeyError:
        # todo: clean up dirty hack
        #  mostly for column_value, which seems to be not documented by orcl anyway
        #  other objects also do not have their own specific entries
        obj_type = api_objects["D2FFO_ANY"]
        object_number = 6
    else:
        object_number = obj_type["object_number"]

    return obj_type, object_number


def add_properties(klass: Type[BaseObject], api_objects: Dict) -> Type[BaseObject]:  # type: ignore
    obj_type, klass._object_number = object_type(klass, api_objects)

    # todo: at this point, forms should be initialized, should be able to dynamically
    #  add properties..?
    for forms_object_property in obj_type["properties"]:

        property_number = forms_object_property["property_number"]

        attribute: Union[str, Union[Common, Subobjects[BaseObject]]]
        # todo: probably do not even need this function call anymore...
        prop_name, attribute = property_attribute(property_number)

        if prop_name and "(obsolete)" not in prop_name:
            try:
                prop_name = Properties(property_number).name.rstrip("_")
            except ValueError:  # pragma: no cover
                raise RuntimeError(f"Unrecognized property ({prop_name})")

            if prop_name in dir(klass):
                continue
            setattr(klass, prop_name, attribute)

    return klass


def forms_object(
    klass: Union[Type[Module], Type[GenericObject]]
) -> Union[Type[Module], Type[GenericObject]]:
    registered_objects[klass.object_type.value[6:]] = klass
    return klass



from .context import context
from .context import move
from .context import property_type
from .context import property_constant_number
from .context import object_number
from .generic_object import FormsObjects
from .error_handling import FormsException
from .forms_api import read_api_objects
from .forms_objects import Alert
from .forms_objects import AttachedLibrary
from .forms_objects import Canvas
from .forms_objects import ColumnValue
from .forms_objects import CompoundText
from .forms_objects import DataBlock
from .forms_objects import DataSourceArgument
from .forms_objects import DataSourceColumn
from .forms_objects import Editor
from .forms_objects import Event
from .forms_objects import FormParameter
from .forms_objects import GenericObject
from .forms_objects import Graphic
from .forms_objects import Item
from .forms_objects import LOV
from .forms_objects import LOVColumnMap
from .forms_objects import Menu
from .forms_objects import MenuItem
from .forms_objects import Module
from .forms_objects import ObjectChild
from .forms_objects import ObjectGroup
from .forms_objects import Point
from .forms_objects import ProgramUnit
from .forms_objects import PropertyClass
from .forms_objects import RadioButton
from .forms_objects import RecordGroup
from .forms_objects import RecordGroupColspec
from .forms_objects import Relation
from .forms_objects import Report
from .forms_objects import TabPage
from .forms_objects import Trigger
from .forms_objects import VisualAttribute
from .forms_objects import VisualState
from .forms_objects import Window
from .misc import add_properties
from .misc import registered_objects
from .property_types import Properties
from .constants import Justification
from .constants import DefaultAlertButton
from .constants import AlertStyle
from .constants import Bevel
from .constants import CaseRestriction
from .constants import CheckBoxMappingofOtherValues
from .constants import CanvasType
from .constants import ColumnDataType
from .constants import CommandType
from .constants import ImageFormat
from .constants import ImageDepth
from .constants import CoordinateSystem
from .constants import CursorMode
from .constants import DeleteRecordBehavior
from .constants import InitialKeyboardState
from .constants import Direction
from .constants import ExecutionHierarchy
from .constants import FontStyle
from .constants import FontWeight
from .constants import FontSpacing
from .constants import ParameterDataType
from .constants import ExecutionMode
from .constants import ItemType
from .constants import DataType
from .constants import KeyMode
from .constants import PLSQLLibrarySource
from .constants import LockingMode
from .constants import ListType
from .constants import ListStyle
from .constants import MagicItem
from .constants import MenuItemType
from .constants import CompressionQuality
from .constants import MouseNavigationLimit
from .constants import NavigationStyle
from .constants import DisplayQuality
from .constants import ReadingOrder
from .constants import RealUnit
from .constants import RecordGroupType
from .constants import RecordOrientation
from .constants import ScrollBarOrientation
from .constants import SizingStyle
from .constants import TriggerStyle
from .constants import PromptFontStyle
from .constants import CalculationMode
from .constants import PromptFontWeight
from .constants import PromptFontSpacing
from .constants import PromptDisplayStyle
from .constants import ValidationUnit
from .constants import WindowStyle
from .constants import WrapStyle
from .constants import GraphicsType
from .constants import ArrowStyle
from .constants import PromptReadingOrder
from .constants import PromptJustification
from .constants import PromptAttachmentEdge
from .constants import PromptAlignment
from .constants import FrameTitleReadingOrder
from .constants import FrameTitleAlignment
from .constants import LayoutStyle
from .constants import FrameAlignment
from .constants import SingleObjectAlignment
from .constants import StartPromptAlignment
from .constants import TopPromptAlignment
from .constants import UpdateLayout
from .constants import DashStyle
from .constants import CapStyle
from .constants import JoinStyle
from .constants import LineSpacing
from .constants import HorizontalOrigin
from .constants import VerticalOrigin
from .constants import HorizontalJustification
from .constants import VerticalJustification
from .constants import SummaryFunction
from .constants import ArgumentMode
from .constants import CommunicationMode
from .constants import ColumnType
from .constants import ArgumentType
from .constants import QueryDataSourceType
from .constants import DMLDataTargetType
from .constants import TabAttachmentEdge
from .constants import VisualAttributeType
from .constants import FrameTitleFontStyle
from .constants import FrameTitleFontWeight
from .constants import FrameTitleFontSpacing
from .constants import CornerStyle
from .constants import InteractionMode
from .constants import IsolationMode
from .constants import ReportDestinationType
from .constants import KeyboardState
from .constants import ScrollBarAlignment
from .constants import ProgramUnitType
from .constants import RuntimeCompatibilityMode
from .constants import GraphicFontWeight
from .constants import GraphicFontSpacing
from .constants import RelationType
from .constants import DataLengthSemantics
from .constants import EventType
from .constants import Scope
from .constants import ViewMode

__version__ = "0.3.7"


def initialize_context(version: str = "12c", encoding: str = "utf-8") -> None:
    context.init(version=version, encoding=encoding)

    api_objects = read_api_objects(version=version)

    for forms_object in registered_objects.values():
        # todo: i could just add static ones for all objects
        #  and still have this code run adding missing values
        #  or by removing ones not present in the version
        if forms_object != Module:
            add_properties(forms_object, api_objects)



from __future__ import annotations

import enum
from ctypes import c_void_p
from typing import Optional, Union, NoReturn

from .context import create
from .context import destroy
from .context import has_property
from .context import remove_subclass
from .context import set_subclass
from .context import move
from .context import query_type
from .context import is_subclassed

from .property_types import Properties


class FormsObjects(enum.Enum):
    canvas = "D2FFO_CANVAS"
    alert = "D2FFO_ALERT"
    attached_library = "D2FFO_ATT_LIB"
    data_block = "D2FFO_BLOCK"
    form_parameter = "D2FFO_FORM_PARAM"
    graphic = "D2FFO_GRAPHIC"
    item = "D2FFO_ITEM"
    point = "D2FFO_POINT"
    program_unit = "D2FFO_PROG_UNIT"
    property_class = "D2FFO_PROP_CLASS"
    radio_button = "D2FFO_RADIO_BUTTON"
    relation = "D2FFO_RELATION"
    tab_page = "D2FFO_TAB_PAGE"
    trigger = "D2FFO_TRIGGER"
    visual_attribute = "D2FFO_VIS_ATTR"
    visual_state = "D2FFO_VIS_STATE"  # todo: undocumented, no .h file
    compound_text = "D2FFO_CMPTXT"  # todo: undocumented, no .h file
    window = "D2FFO_WINDOW"
    module = "D2FFO_FORM_MODULE"
    data_source_argument = "D2FFO_DAT_SRC_ARG"
    data_source_column = "D2FFO_DAT_SRC_COL"
    editor = "D2FFO_EDITOR"
    event = "D2FFO_EVENT"
    lov = "D2FFO_LOV"
    lov_column_map = "D2FFO_LV_COLMAP"
    menu = "D2FFO_MENU"
    menu_item = "D2FFO_MENU_ITEM"
    object_group = "D2FFO_OBJ_GROUP"
    object_child = "D2FFO_OBG_CHILD"
    record_group = "D2FFO_REC_GROUP"
    record_group_colspec = "D2FFO_RG_COLSPEC"
    report = "D2FFO_REPORT"
    column_value = "D2FFO_COLUMN_VALUE"


class ValueTypes(enum.IntEnum):
    UNKNOWN = 0
    BOOLEAN = 1
    NUMBER = 2
    TEXT = 3
    OBJECT = 4


PropertyTypes = Union[bool, int, str, bytes, "BaseObject", c_void_p]


class BaseObject:
    object_type: FormsObjects
    _object_number: Optional[int]
    _as_parameter_: c_void_p

    def __init__(
        self, generic_object: Union[c_void_p, BaseObject, GenericObject]
    ) -> None:
        if isinstance(generic_object, BaseObject):
            self._as_parameter_ = generic_object._as_parameter_
        else:
            self._as_parameter_ = generic_object

    def has_property(self, property_number: int) -> bool:
        return has_property(self, property_number)

    def destroy(self) -> None:
        destroy(self)
        self._as_parameter_ = c_void_p(0)

    def remove_subclass(self) -> None:
        remove_subclass(self)

    def set_subclass(self, property_class: BaseObject, keep_path: bool = False) -> None:
        set_subclass(self, property_class, keep_path)

    def move(self, next_object: GenericObject) -> None:
        move(self, next_object)

    def query_type(self) -> int:
        return query_type(self)

    def duplicate(self, new_owner: GenericObject, new_name: str) -> NoReturn:
        raise NotImplementedError()

    def replicate(self, new_owner: GenericObject, new_name: str) -> NoReturn:
        raise NotImplementedError()

    def find_object(self, name: str, object_type: FormsObjects) -> NoReturn:
        raise NotImplementedError()

    def inherit_property(self, property_type: Properties) -> NoReturn:
        raise NotImplementedError()

    def is_property_inherited(self, property_type: Properties) -> NoReturn:
        raise NotImplementedError()

    def is_property_default(self, property_type: Properties) -> NoReturn:
        raise NotImplementedError()

    def is_subclassed(self) -> bool:
        return is_subclassed(self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self._as_parameter_)})"

    def __bool__(self) -> bool:
        return bool(self._as_parameter_)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseObject):
            return NotImplemented
        return self._as_parameter_ == other._as_parameter_


class GenericObject(BaseObject):
    @classmethod
    def create(cls, owner: BaseObject, name: str) -> GenericObject:
        # todo: maybe better way than just an assert?
        assert cls._object_number is not None
        return cls(create(owner, name, cls._object_number))



import enum


class Properties(enum.IntEnum):  # pragma: no cover
    keyboard_accelerator = 2
    alerts = 3
    justification = 4
    button_1_label = 5
    button_2_label = 6
    button_3_label = 7
    default_alert_button = 8
    message = 9
    alert_style = 10
    attached_libraries = 11
    automatic_select = 12
    automatic_display = 13
    automatic_query = 14
    automatic_refresh = 15
    automatic_skip = 16
    background_color = 17
    database_item = 18
    bevel = 19
    mirror_item_object = 20
    data_blocks = 21
    bottom_title = 22
    graphics = 23
    canvases = 24
    canvas = 25
    canvas_object_pointer = 26
    case_info = 27
    case_insensitive_query = 28
    case_restriction = 29
    check_box_mapping_of_other_values = 30
    value_when_checked = 33
    close_allowed = 34
    canvas_type = 35
    synchronize_with_item = 36
    column_data_type = 37
    column_mapping_object = 38
    enforce_column_security = 39
    record_group_colspecs = 40
    command_text = 41
    command_type = 42
    comments = 43
    image_format = 44
    console_window = 45
    image_depth = 46
    cursor_mode = 48
    deferred = 49
    delete_allowed = 50
    delete_record_behavior = 51
    prompt_foreground_color = 52
    detail_data_block = 53
    default_button = 54
    initial_keyboard_state = 55
    direction = 57
    visible = 58
    display_without_privilege = 59
    display_viewport = 60
    display_width = 61
    viewport_x_position = 62
    viewport_y_position = 63
    editors = 64
    editor = 65
    editor_object_pointer = 66
    editor_x_position = 67
    editor_y_position = 68
    enabled = 69
    execution_hierarchy = 70
    fill_pattern = 73
    fire_in_enter_query_mode = 74
    first_navigation_data_block = 75
    first_data_block_object = 76
    resize_allowed = 78
    menu_module = 79
    menu_role = 80
    initial_menu = 81
    font_name = 82
    font_size = 83
    font_style = 84
    font_weight = 85
    font_spacing = 86
    foreground_color = 87
    format_mask = 88
    form_parameters = 89
    parameter_data_type = 90
    parameter_initial_value = 91
    filename = 92
    execution_mode = 93
    height = 94
    highest_allowed_value = 95
    hint = 96
    show_horizontal_scroll_bar = 97
    horizontal_toolbar_canvas = 98
    iconic = 99
    minimize_allowed = 100
    icon_filename = 101
    minimized_title = 102
    inherit_menu = 103
    insert_allowed = 104
    items = 105
    item_type = 106
    copy_value_from_item = 107
    data_type = 108
    initial_value = 109
    join_condition = 115
    keep_cursor_position = 116
    key_mode = 117
    label = 118
    prompt_background_color = 119
    pl_sql_library_location = 120
    pl_sql_library_source = 121
    locking_mode = 123
    lock_record = 124
    filter_before_display = 125
    lovs = 126
    list_of_values = 127
    lov_object_pointer = 128
    list_type = 129
    validate_from_list = 130
    list_x_position = 131
    list_y_position = 132
    lowest_allowed_value = 133
    list_style = 134
    magic_item = 135
    main_menu = 136
    maximum_length = 137
    menus = 138
    form_horizontal_toolbar_canvas = 139
    form_vertical_toolbar_canvas = 140
    menu_directory = 141
    menu_filename = 142
    menu_item_object = 143
    menu_item_type = 144
    compression_quality = 145
    associated_menus = 146
    menu_item_radio_group = 147
    access_key = 148
    modal = 149
    mouse_navigation_limit = 150
    mouse_navigate = 151
    move_allowed = 152
    multi_line = 153
    name_ = 154  # fixme: conflicts with enum.IntEnum...
    prompt_fill_pattern = 155
    keyboard_navigable = 156
    navigation_style = 157
    next_object = 159
    next_navigation_data_block = 160
    next_data_block_object = 161
    next_navigation_item = 162
    next_item_object = 163
    object_groups = 164
    object_group_child_object = 165
    old_lov_text = 166
    optimizer_hint = 175
    order_by_clause = 176
    mapping_of_other_values = 177
    primary_key = 178
    program_units = 179
    property_classes = 180
    prevent_masterless_operations = 181
    previous_navigation_data_block = 182
    previous_data_block_object = 183
    previous_navigation_item = 184
    previous_item_object = 185
    query_allowed = 186
    query_length = 187
    query_only = 188
    display_quality = 189
    radio_button_object = 190
    raise_on_entry = 191
    radio_button_value = 192
    reading_order = 193
    number_of_records_buffered = 195
    number_of_records_displayed = 196
    query_array_size = 197
    distance_between_records = 198
    record_groups = 199
    record_group = 200
    record_group_object_pointer = 201
    record_group_query = 202
    record_group_type = 203
    record_orientation = 204
    current_record_visual_attribute_group = 205
    relation_object = 206
    hide_on_exit = 207
    rendered = 208
    required = 209
    return_item = 210
    reverse_direction = 211
    hide_object = 213
    prompt_font_name = 214
    savepoint_mode = 215
    scroll_bar_canvas = 216
    scroll_bar_orientation = 217
    scroll_bar_width = 218
    scroll_bar_x_position = 219
    scroll_bar_y_position = 220
    show_scroll_bar = 221
    conceal_data = 222
    display_in_keyboard_help = 223
    sizing_style = 224
    startup_code = 225
    sub_title = 226
    tear_off_menu = 229
    title = 230
    prompt_font_size = 231
    triggers = 232
    keyboard_help_text = 233
    trigger_step_object = 234
    trigger_style = 236
    trigger_text = 237
    value_when_unchecked = 238
    update_allowed = 239
    update_changed_columns_only = 240
    update_only_if_null = 241
    use_3d_controls = 242
    use_security = 243
    visual_attribute_group = 245
    va_object = 246
    prompt_font_style = 247
    calculation_mode = 248
    prompt_font_weight = 249
    prompt_font_spacing = 250
    column_name = 251
    prompt_display_style = 252
    show_vertical_scroll_bar = 253
    viewport_height = 254
    viewport_width = 255
    viewport_x_position_on_canvas = 256
    viewport_y_position_on_canvas = 257
    visual_attributes = 258
    visual_stat = 259
    validation_unit = 260
    vertical_toolbar_canvas = 261
    where_clause = 262
    width = 263
    windows = 264
    enforce_primary_key = 265
    window = 268
    window_object_pointer = 269
    window_style = 270
    primary_canvas = 271
    wrap_style = 272
    x_position = 273
    y_position = 274
    maximize_allowed = 275
    display_hint_automatically = 277
    tooltip = 279
    current_record_va_pointer = 280
    graphics_type = 281
    closed = 282
    arrow_style = 283
    x_corner_radius = 284
    y_corner_radius = 285
    start_angle = 286
    end_angle = 287
    prompt = 288
    prompt_reading_order = 289
    prompt_justification = 290
    previous_object = 291
    prompt_attachment_edge = 292
    prompt_attachment_offset = 293
    prompt_alignment = 294
    prompt_alignment_offset = 295
    owning_object = 296
    dither = 297
    clip_x_position = 298
    clip_y_position = 299
    clip_width = 300
    clip_height = 301
    points = 302
    frame_title_reading_order = 303
    frame_title_alignment = 304
    frame_title_offset = 305
    frame_title_spacing = 306
    frame_title = 307
    layout_style = 308
    frame_alignment = 309
    single_object_alignment = 310
    vertical_fill = 311
    maximum_objects_per_line = 312
    allow_expansion = 313
    horizontal_margin = 314
    vertical_margin = 315
    vertical_object_offset = 316
    horizontal_object_offset = 317
    start_prompt_offset = 318
    top_prompt_offset = 319
    start_prompt_alignment = 320
    top_prompt_alignment = 321
    allow_multi_line_prompts = 322
    allow_top_attached_prompts = 323
    allow_start_attached_prompts = 324
    update_layout = 325
    line_width = 326
    rotation_angle = 327
    edge_foreground_color = 328
    edge_background_color = 329
    edge_pattern = 330
    dash_style = 331
    cap_style = 332
    join_style = 333
    line_spacing = 334
    custom_spacing = 335
    fixed_bounding_box = 336
    wrap_text = 337
    bounding_box_scaleable = 338
    font_scaleable = 339
    horizontal_origin = 340
    vertical_origin = 341
    horizontal_justification = 342
    vertical_justification = 343
    compound_text_object = 344
    formula = 346
    summary_function = 347
    summarized_item = 348
    number_of_items_displayed = 349
    scrollbar_canvas_object_pointer = 350
    argument_mode = 351
    tab_pages = 356
    tab_page = 357
    tab_page_name = 358
    communication_mode = 359
    data_source_data_block = 360
    data_source_x_axis = 361
    data_source_y_axis = 362
    update_on_commit = 363
    update_on_query = 364
    query_name = 365
    popup_menu = 366
    popup_menu_object = 367
    dml_data_target_name = 370
    query_data_source_name = 371
    data_source_column_name = 372
    data_source_column_type = 373
    argument_name = 374
    argument_type = 375
    argument_value = 376
    database_data_block = 377
    query_data_source_type = 378
    query_all_records = 379
    dml_data_target_type = 380
    dml_array_size = 381
    tooltip_visual_attribute_group = 382
    tooltip_va_object = 383
    insert_procedure_name = 419
    update_procedure_name = 422
    delete_procedure_name = 425
    lock_procedure_name = 428
    single_record = 431
    trigger_internal_type = 432
    length = 433
    mandatory = 434
    precision = 435
    scale = 436
    prompt_va_object = 437
    prompt_visual_attribute_group = 438
    layout_data_block = 439
    owning_module = 440
    source_object = 441
    scroll_bar_tab_page = 442
    scrollbar_tab_object = 443
    visible_in_vertical_menu_toolbar = 444
    visible_in_horizontal_menu_toolbar = 445
    tab_attachment_edge = 446
    argument_type_name = 447
    visible_in_menu = 448
    visual_attribute_type = 450
    program_unit_text = 451
    icon_in_menu = 452
    automatic_position = 453
    automatic_column_width = 454
    record_group_fetch_size = 455
    title_va_object = 456
    frame_title_visual_attribute_group = 457
    frame_title_foreground_color = 458
    frame_title_background_color = 459
    frame_title_fill_pattern = 460
    frame_title_font_name = 461
    frame_title_font_size = 462
    frame_title_font_style = 463
    frame_title_font_weight = 464
    frame_title_font_spacing = 465
    corner_style = 466
    interaction_mode = 467
    maximum_query_time = 468
    maximum_records_fetched = 469
    isolation_mode = 470
    count_of_data_items_in_the_column = 471
    real_object_pointed_to_by_the_object_group_child = 472
    reports = 473
    count_of_list_element_items = 474
    precompute_summaries = 475
    report_destination_type = 476
    report_destination_name = 477
    report_destination_format = 478
    report_server = 479
    other_reports_parameters = 480
    share_library_with_form = 487
    keyboard_state = 488
    summarized_block = 490
    scroll_bar_alignment = 491
    shrinkwrap = 492
    graphic_object_font_name = 493
    graphic_object_font_size = 494
    graphic_object_font_style = 495
    graphic_object_text = 496
    parent_objects_module = 497
    parent_objects_module_type = 498
    parent_object_name = 499
    parent_objects_file_name = 500
    parent_objects_file_path = 501
    count_of_roles = 503
    program_unit_type = 504
    submenu_name = 505
    submenu_object = 506
    menu_item_code = 507
    query_column_object = 508
    query_argument_object = 509
    insert_column_object = 510
    insert_argument_object = 511
    update_column_object = 512
    update_argument_object = 513
    delete_column_object = 514
    delete_argument_object = 515
    lock_column_object = 516
    lock_argument_object = 517
    runtime_compatibility_mode = 532
    object_group_type = 533
    graphic_font_weight = 535
    graphic_font_spacing = 536
    parent_objects_first_level_owners_name = 537
    parent_objects_first_level_owners_type = 538
    parent_objects_second_level_owners_name = 539
    parent_objects_second_level_owners_type = 540
    parent_objects_type = 541
    scroll_bar_length = 542
    relation_type = 543
    detail_reference_item = 544
    column_type_name = 545
    parent_column = 546
    alias = 547
    include_ref_item = 548
    help_book_title = 550
    help_book_topic = 551
    allow_empty_branches = 552
    multi_selection = 553
    show_lines = 554
    show_symbols = 555
    tre_record_group = 556
    data_query = 557
    implementation_class = 558
    dml_returning_value = 559
    persistent_client_info_storage = 560
    persistent_client_info_storage_length = 561
    active_style = 562
    width_style = 563
    column_value = 570
    defer_required_enforcement = 572
    data_length_semantics = 573
    event = 574
    event_type = 575
    subscription_name = 576
    scope = 577
    auto_subscribe = 578
    event_implementation_class = 579
    correlation_id = 580
    priority_mode = 581
    view_mode = 582
    report_object_type = 583
    delivery_type = 584
    ssl_connection = 585
    service_location = 586
    report_path = 587
    report_locale = 588
    report_format = 589
    template_name = 590
    report_parameters = 591
    ftp_server_name = 592
    ftp_user = 593
    ftp_file_name = 594
    secure_ftp = 595
    number = 596
    fax_server_name = 597
    to_email = 598
    from_email = 599
    cc_email = 600
    bcc_email = 601
    reply_to = 602
    subject = 603
    body = 604
    mail_server_name = 605
    local_file_name = 606
    number_of_copies = 607
    orientation = 608
    page_range = 609
    side = 610
    tray = 611
    print_server_name = 612
    webdav_auth_type = 613
    webdav_user = 614
    webdav_file_path = 615
    webdav_server_name = 616
    row_banding_frequency = 617
    cursor_style = 618
    gradient_start_side = 619



import setuptools

from pyoracle_forms import __version__

with open("README.rst", mode="r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyoracle_forms",
    version=__version__,
    author="LatvianPython",
    author_email="kalvans.rolands@gmail.com",
    description="Python wrapper of the Oracle Forms API",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/LatvianPython/pyoracle_forms",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Development Status :: 2 - Pre-Alpha",
    ],
    include_package_data=True,
    python_requires=">=3.9",
)



import enum


class Justification(enum.IntEnum):
    left = 0
    right = 1
    center = 2
    start = 3
    end = 4


class DefaultAlertButton(enum.IntEnum):
    button_1 = 0
    button_2 = 1
    button_3 = 2


class AlertStyle(enum.IntEnum):
    stop = 0
    caution = 1
    note = 2


class Bevel(enum.IntEnum):
    raised = 0
    lowered = 1
    none = 2
    inset = 3
    outset = 4
    plain = 5


class CaseRestriction(enum.IntEnum):
    mixed = 0
    upper = 1
    lower = 2


class CheckBoxMappingofOtherValues(enum.IntEnum):
    not_allowed = 0
    checked = 1
    unchecked = 2


class CanvasType(enum.IntEnum):
    content = 0
    stacked = 1
    vertical_toolbar = 2
    horizontal_toolbar = 3
    tab = 4


class ColumnDataType(enum.IntEnum):
    character = 0
    number = 1
    date = 2
    long = 3
    object_ref = 4


class CommandType(enum.IntEnum):
    null = 0
    menu = 1
    plsql = 2


class ImageFormat(enum.IntEnum):
    bmp = 0
    cals = 1
    gif = 2
    jfif = 3
    pict = 4
    ras = 5
    tiff = 6
    tpic = 7
    native = 8


class ImageDepth(enum.IntEnum):
    original = 0
    monochrome = 1
    gray = 2
    lut = 3
    rgb = 4


class CoordinateSystem(enum.IntEnum):
    character = 0
    real = 1


class CursorMode(enum.IntEnum):
    open = 0
    close = 1


class DeleteRecordBehavior(enum.IntEnum):
    cascading = 0
    isolated = 1
    non_isolated = 2


class InitialKeyboardState(enum.IntEnum):
    default = 0
    roman = 1
    local = 2


class Direction(enum.IntEnum):
    default = 0
    left_to_right = 1
    right_to_left = 2


class ExecutionHierarchy(enum.IntEnum):
    override = 0
    before = 1
    after = 2


class FontStyle(enum.IntEnum):
    plain = 0
    italic = 1
    oblique = 2
    underline = 3
    outline = 4
    shadow = 5
    inverted = 6
    overstrike = 7
    blink = 8


class FontWeight(enum.IntEnum):
    ultralight = 0
    extralight = 1
    light = 2
    demilight = 3
    medium = 4
    demibold = 5
    bold = 6
    extrabold = 7
    ultrabold = 8


class FontSpacing(enum.IntEnum):
    ultradense = 0
    extradense = 1
    dense = 2
    semidense = 3
    normal = 4
    semiexpand = 5
    expand = 6
    extraexpand = 7
    ultraexpand = 8


class ParameterDataType(enum.IntEnum):
    char = 0
    number = 1
    date = 2


class ExecutionMode(enum.IntEnum):
    batch = 0
    runtime = 1


class ItemType(enum.IntEnum):
    bean_area = 0
    check_box = 1
    display_item = 2
    hierarchical_tree = 3
    image = 4
    list_item = 5
    push_button = 6
    radio_group = 7
    text_item = 8
    user_area = 9
    activex_control_obsolete = 10
    chart_item_obsolete = 11
    ole_container_obsolete = 12
    sound_obsolete = 13
    vbx_control_obsolete = 14


class DataType(enum.IntEnum):
    char = 0
    number = 1
    date = 2
    alpha = 3
    integer = 4
    datetime = 5
    long = 6
    rnumber = 7
    jdate = 8
    edate = 9
    time = 10
    rinteger = 11
    money = 12
    rmoney = 13
    object_ref = 14
    nchar = 15


class KeyMode(enum.IntEnum):
    unique = 0
    updateable = 1
    non_updateable = 2
    automatic = 3


class PLSQLLibrarySource(enum.IntEnum):
    file = 0
    database = 1


class LockingMode(enum.IntEnum):
    immediate = 0
    delayed = 1
    automatic = 2


class ListType(enum.IntEnum):
    record_group = 0
    old = 1


class ListStyle(enum.IntEnum):
    poplist = 0
    tlist = 1
    combo_box = 2


class MagicItem(enum.IntEnum):
    none = 0
    cut = 1
    copy = 2
    paste = 3
    clear = 4
    undo = 5
    help = 6
    about = 7
    quit = 8
    window = 9
    page_setup_obsolete = 10


class MenuItemType(enum.IntEnum):
    plain = 0
    check = 1
    radio = 2
    separator = 3
    magic = 4


class CompressionQuality(enum.IntEnum):
    none = 0
    minimum = 1
    low = 2
    medium = 3
    high = 4
    maximum = 5


class MouseNavigationLimit(enum.IntEnum):
    form = 0
    data_block = 1
    record = 2
    item = 3


class NavigationStyle(enum.IntEnum):
    same_record = 0
    change_record = 1
    change_data_block = 2


class DisplayQuality(enum.IntEnum):
    high = 0
    medium = 1
    low = 2


class ReadingOrder(enum.IntEnum):
    default = 0
    left_to_right = 1
    right_to_left = 2


class RealUnit(enum.IntEnum):
    pixel = 0
    centimeter = 1
    inch = 2
    point = 3
    decipoint = 4


class RecordGroupType(enum.IntEnum):
    query = 0
    static = 1


class RecordOrientation(enum.IntEnum):
    vertical = 0
    horizontal = 1


class ScrollBarOrientation(enum.IntEnum):
    vertical = 0
    horizontal = 1


class SizingStyle(enum.IntEnum):
    crop = 0
    adjust = 1
    fill = 2


class TriggerStyle(enum.IntEnum):
    plsql = 0
    v2 = 1


class PromptFontStyle(enum.IntEnum):
    plain = 0
    italic = 1
    oblique = 2
    underline = 3
    outline = 4
    shadow = 5
    inverted = 6
    overstrike = 7
    blink = 8


class CalculationMode(enum.IntEnum):
    none = 0
    formula = 1
    summary = 2


class PromptFontWeight(enum.IntEnum):
    ultralight = 0
    extralight = 1
    light = 2
    demilight = 3
    medium = 4
    demibold = 5
    bold = 6
    extrabold = 7
    ultrabold = 8


class PromptFontSpacing(enum.IntEnum):
    ultradense = 0
    extradense = 1
    dense = 2
    semidense = 3
    normal = 4
    semiexpand = 5
    expand = 6
    extraexpand = 7
    ultraexpand = 8


class PromptDisplayStyle(enum.IntEnum):
    hidden = 0
    first_record = 1
    all_records = 2


class ValidationUnit(enum.IntEnum):
    default = 0
    form = 1
    data_block = 2
    record = 3
    item = 4


class WindowStyle(enum.IntEnum):
    document = 0
    dialog = 1


class WrapStyle(enum.IntEnum):
    none = 0
    character = 1
    word = 2


class GraphicsType(enum.IntEnum):
    arc = 0
    image = 1
    line = 2
    polygon = 3
    rectangle = 4
    rounded_rectangle = 5
    text = 6
    group = 7
    frame = 8


class ArrowStyle(enum.IntEnum):
    none = 0
    start = 1
    end = 2
    both_ends = 3
    middle_to_start = 4
    middle_to_end = 5


class PromptReadingOrder(enum.IntEnum):
    default = 0
    left_to_right = 1
    right_to_left = 2


class PromptJustification(enum.IntEnum):
    left = 0
    right = 1
    center = 2
    start = 3
    end = 4


class PromptAttachmentEdge(enum.IntEnum):
    start = 0
    end = 1
    top = 2
    bottom = 3


class PromptAlignment(enum.IntEnum):
    start = 0
    end = 1
    center = 2


class FrameTitleReadingOrder(enum.IntEnum):
    default = 0
    left_to_right = 1
    right_to_left = 2


class FrameTitleAlignment(enum.IntEnum):
    left = 0
    right = 1
    center = 2
    start = 3
    end = 4


class LayoutStyle(enum.IntEnum):
    form = 0
    tabular = 1


class FrameAlignment(enum.IntEnum):
    start = 0
    end = 1
    center = 2
    fill = 3
    column = 4


class SingleObjectAlignment(enum.IntEnum):
    start = 0
    end = 1
    center = 2


class StartPromptAlignment(enum.IntEnum):
    start = 0
    end = 1
    center = 2


class TopPromptAlignment(enum.IntEnum):
    start = 0
    end = 1
    center = 2


class UpdateLayout(enum.IntEnum):
    manually = 0
    automatically = 1
    locked = 2


class DashStyle(enum.IntEnum):
    solid = 0
    dotted = 1
    dashed = 2
    dash_dot = 3
    double_dot = 4
    long_dash = 5
    dash_double_dot = 6


class CapStyle(enum.IntEnum):
    butt = 0
    round = 1
    projecting = 2


class JoinStyle(enum.IntEnum):
    mitre = 0
    bevel = 1
    round = 2


class LineSpacing(enum.IntEnum):
    single = 0
    one_and_a_half = 1
    double = 2
    custom = 3


class HorizontalOrigin(enum.IntEnum):
    left = 0
    right = 1
    center = 2


class VerticalOrigin(enum.IntEnum):
    top = 0
    center = 1
    bottom = 2


class HorizontalJustification(enum.IntEnum):
    left = 0
    right = 1
    center = 2
    start = 3
    end = 4


class VerticalJustification(enum.IntEnum):
    top = 0
    center = 1
    bottom = 2


class SummaryFunction(enum.IntEnum):
    none = 0
    avg = 1
    count = 2
    max = 3
    min = 4
    stddev = 5
    sum = 6
    variance = 7


class ArgumentMode(enum.IntEnum):
    in_ = 0
    out = 1
    in_out = 2


class CommunicationMode(enum.IntEnum):
    synchronous = 0
    asynchronous = 1


class ColumnType(enum.IntEnum):
    varchar2 = 0
    number = 1
    long = 2
    rowid = 3
    date = 4
    raw = 5
    long_raw = 6
    char = 7
    mlslabel = 8
    table = 9
    record = 10
    refcursor = 11
    named_type = 12
    object_ref = 13
    varray = 14
    nested_table = 15
    blob = 16
    clob = 17
    bfile = 18
    nvarchar2 = 19
    nchar = 20
    nclob = 21


class ArgumentType(enum.IntEnum):
    varchar2 = 0
    number = 1
    long = 2
    rowid = 3
    date = 4
    raw = 5
    long_raw = 6
    char = 7
    mlslabel = 8
    table = 9
    record = 10
    refcursor = 11
    named_type = 12
    object_ref = 13
    varray = 14
    nested_table = 15
    blob = 16
    clob = 17
    bfile = 18
    nvarchar2 = 19
    nchar = 20
    nclob = 21


class QueryDataSourceType(enum.IntEnum):
    none = 0
    table = 1
    procedure = 2
    transactional_triggers = 3
    from_clause_query = 4


class DMLDataTargetType(enum.IntEnum):
    none = 0
    table = 1
    procedure = 2
    transactional_triggers = 3


class TabAttachmentEdge(enum.IntEnum):
    top = 0
    bottom = 1
    left = 2
    right = 3
    start = 4
    end = 5


class VisualAttributeType(enum.IntEnum):
    common = 0
    prompt = 1
    title = 2


class FrameTitleFontStyle(enum.IntEnum):
    plain = 0
    italic = 1
    oblique = 2
    underline = 3
    outline = 4
    shadow = 5
    inverted = 6
    overstrike = 7
    blink = 8


class FrameTitleFontWeight(enum.IntEnum):
    ultralight = 0
    extralight = 1
    light = 2
    demilight = 3
    medium = 4
    demibold = 5
    bold = 6
    extrabold = 7
    ultrabold = 8


class FrameTitleFontSpacing(enum.IntEnum):
    ultradense = 0
    extradense = 1
    dense = 2
    semidense = 3
    normal = 4
    semiexpand = 5
    expand = 6
    extraexpand = 7
    ultraexpand = 8


class CornerStyle(enum.IntEnum):
    chamfered = 0
    square = 1
    rounded = 2


class InteractionMode(enum.IntEnum):
    blocking = 0
    non_blocking = 1


class IsolationMode(enum.IntEnum):
    read_committed = 0
    serializable = 1


class ReportDestinationType(enum.IntEnum):
    preview = 0
    file = 1
    printer = 2
    mail = 3
    cache = 4
    screen = 5


class KeyboardState(enum.IntEnum):
    any = 0
    roman_only = 1
    local_only = 2


class ScrollBarAlignment(enum.IntEnum):
    start = 0
    end = 1


class ProgramUnitType(enum.IntEnum):
    unknown = 0
    procedure = 1
    function = 2
    package_spec = 3
    package_body = 4


class RuntimeCompatibilityMode(enum.IntEnum):
    v4_5 = 0
    v5_0 = 1


class GraphicFontWeight(enum.IntEnum):
    ultralight = 0
    extralight = 1
    light = 2
    demilight = 3
    medium = 4
    demibold = 5
    bold = 6
    extrabold = 7
    ultrabold = 8


class GraphicFontSpacing(enum.IntEnum):
    ultradense = 0
    extradense = 1
    dense = 2
    semidense = 3
    normal = 4
    semiexpand = 5
    expand = 6
    extraexpand = 7
    ultraexpand = 8


class RelationType(enum.IntEnum):
    join = 0
    ref = 1


class DataLengthSemantics(enum.IntEnum):
    null = 0
    byte = 1
    char = 2


class EventType(enum.IntEnum):
    database = 0
    user_defined = 1
    system_client_idle = 2
    system_db_idle = 3
    system_single_sign_off = 4
    system_notification = 5
    system_media_completion = 6


class Scope(enum.IntEnum):
    application = 0
    form = 1


class ViewMode(enum.IntEnum):
    browse = 0
    locked = 1
    removed = 2
